; Programa: Package con agentes jugadores simples de Tron,
;           juego tambin denominado "light cycle",
;           ver http://en.wikipedia.org/wiki/Light_cycle
;
; Asignatura: Introduccin a la Inteligencia Artificial
;             ETSII - ULL
; Prctica: (sistemas de produccin)
; Profesor: Patricio Garca (patricio AT etsii.ull.es)
; Fecha:    12/04/2006
; Comentarios:
;
;           Los agentes definidos utilizan el entorno de juego
;           Tron, definido principalmente por los packages
;           tron-base y tron-comm.
;
;           Las funciones accesibles ms importantes son:
;
;             agente-mitad
;             agente-temerario
;             agente-espiral
;             agente-hormiga-borracha
;
;           Ver la documentacin existente para cada una de ellas.
;
;           Para su utilizacin es aconsejable usar el cargador
;           del entorno de juego tron: tron-loader.lsp.
;
;
; Dependencias:
;
;           Hace uso del paquete que contiene la interface (tron-comm)
;           con el control del juego (tron-base).
;
;
; Modificaciones:
;
;           - Ver 0.1 (13/04/2006): Primera versin. Extraida
;             de tron.0.1.lsp. Vease tron-base.lsp y tron-comm*.lsp
;
;           - Ver 0.2 (21/04/2006): Revisados nombre de variables y
;             funciones. Transformado todo a funciones recursivas.
;             Algunas simplificaciones del cdigo. Ampliada
;             documentacin.
;
;           - Ver 0.3 (29/04/2006): Empaquetado, puesto interface
;             de definicin de agentes (const agentes). Ofuscado
;             cdigo de dir-vacia-pref.
;
;           - Ver 0.3.1 (02/04/2006): Incluido agente mitad.
;
;           - Ver 0.3.2 (05/04/2006): Cambiado nombre constante
;             *version-tron-agentes-simples*.
;
;
; Por hacer:
;
;
;


;
; Definicin del package
; ----------------------

(defpackage "TRON-AGENTES-SIMPLES"
  (:nicknames "TR-AGENTES-SIMPLES" "TRON-A-S")
  (:use "LISP" "TRON-COMM")
  (:export "AGENTES" "*VERSION-TRON-AGENTES-SIMPLES*" "AGENTE-MITAD" "AGENTE-TEMERARIO" "AGENTE-ESPIRAL" "AGENTE-HORMIGA-BORRACHA" "AGENTE-IMITADOR" "DIR-VACIA-P" "DIR-INC" "DIR-VACIA-PREF"))


(in-package "TRON-AGENTES-SIMPLES")


;
; Constante con lista de agentes presentes en este package
; --------------------------------------------------------

(defconstant agentes
  '((agente-mitad) (agente-temerario) (agente-espiral) (agente-hormiga-borracha) (agente-imitador))
  "Agentes simples para jugar a TRON.")


;
; Constante versin
; -----------------

(defconstant *version-tron-agentes-simples* "0.3.2"
  "Versin del package agentes-simples.")


;
; Funciones de tratamiento de lista de direcciones preferentes
; ------------------------------------------------------------

(defun dir-vacia-p (pos dir)
  "Mira en la posicin POS si movndose en la direccin DIR est vaca, devolviendo DIR en tal caso y NIL en caso contrario."
  (if (null (consulta-pos-tab pos dir))
      dir))

(defun dir-inc (dir dir-ini)
  "Calcula la direccin DIR relativa a DIR-INI, esto es, gira DIR-INI pasos en sentido agujas del reloj sobre DIR."
  (rem (+ dir dir-ini) 4))

(defun dir-vacia-pref (pos lis-dir-pref &optional (dir-ini 0))
  "Busca la primera direccin vaca alrededor de POS siguiendo el orden de LIS-DIR-PREF, si DIR-INI est instanciado las direcciones de LIS-DIR-PREF son relativas a DIR-INI."
  (if (consp lis-dir-pref)
      (or (dir-vacia-p pos (dir-inc (first lis-dir-pref) dir-ini))
	  (dir-vacia-pref pos (rest lis-dir-pref) dir-ini))))


;
; Agente NSEO
; -----------

(defun agente-mitad (id)
  "Agente jugador Tron que slo busca la mitad."
  (dir-vacia-pref (consulta-pos-jug id) '(0 2 1 3)))


;
; Agente rombo derecha
; --------------------

(defun agente-temerario (id)
  "Agente jugador Tron que va temerariamente desenrrollndose!"
  (dir-vacia-pref (consulta-pos-jug id) '(1 3 0) (consulta-est-jug id)))


;
; Agente espiral derecha
; ----------------------
  
(defun agente-espiral (id)
  "Agente jugador Tron que da vueltas en espiral desenrrollndose!"
  (dir-vacia-pref (consulta-pos-jug id) '(1 0 3) (consulta-est-jug id)))


;
; Agente hormiga borracha
; -----------------------

(defvar random-state-hormiga-borracha (make-random-state *random-state*)
  "Estado random para el agente-hormiga-borracha.")

(defun agente-hormiga-borracha (id)
  "Agente jugador Tron que da tumbos y se agarra a las farolas. Elige una direccin aleatorias entre las vacas a su alrededor."
  (let* ((pos (consulta-pos-jug id))
	 (lis-dir (remove nil (mapcar (lambda (dir) (dir-vacia-p pos dir))
				       '(0 1 2 3)))))
    (if lis-dir
	(nth (random (length lis-dir) random-state-hormiga-borracha)
	     lis-dir))))


;
; Agente imitador
; ---------------

(defun agente-imitador (id)
  "Agente jugador Tron que trata de hacer el mismo movimiento que el adversario."
  (dir-vacia-pref (consulta-pos-jug id) 
		  '(0 1 3 2) 
		  (consulta-est-jug (- 3 id))))

