; Programa: Package con agentes jugadores humanos de Tron,
;           juego tambin denominado "light cycle",
;           ver http://en.wikipedia.org/wiki/Light_cycle
;
; Asignatura: Introduccin a la Inteligencia Artificial
;             ETSII - ULL
; Prctica: (sistemas de produccin)
; Profesor: Patricio Garca (patricio AT etsii.ull.es)
; Fecha:    12/04/2006
; Comentarios:
;
;           El agente definido utiliza el entorno de juego
;           Tron, definido principalmente por los packages
;           tron-base y tron-comm.
;
;           Las funciones accesibles ms importantes son:
;
;             agente-humano
;             agente-humano-ini
;             agente-humano-fin
;
;           Ver la documentacin existente para cada una de ellas.
;
;           Para su utilizacin es aconsejable usar el cargador
;           del entorno de juego tron: tron-loader.lsp.
;
;
; Dependencias:
;
;           Hace uso del paquete que contiene la interface (tron-comm)
;           con el control del juego (tron-base).
;
;           Tambin resulta prcticamente intil si no se utiliza
;           junto a un agente visualizador, ver tron-vis-*.lsp.
;
;
;           Debido a la necesidad de controlar el teclado
;           por parte de los jugadores humanos este entorno
;           tiene limitaciones en gcl, de momento (ver bug)
;
;           Tambin tiene problemas con clisp en versiones < 2.33
;           debido a bug en la funcin read-char-no-hang (hace hang),
;           ver: http://groups.google.com/group/comp.lang.lisp/browse_frm/thread/6bfac6923a4126b1/5575c9088e5ca9c3?tvc=1
;
;
; Modificaciones:
;
;           - Ver 0.1 (13/04/2006): Primera versin. Extraida
;             de tron.0.1.lsp. Vease tron-base.lsp y tron-comm*.lsp
;
;           - Ver 0.2 (16/04/2006): Generada versin para clisp,
;             ver echo-hagn-off-clisp.lsp. Arreglado problemas de
;             comido de teclas en juego de dos agentes humanos (ver
;             unread-char-no-hang-no-echo). Arreglado problema de
;             reactivacin del echo en CMU lisp.
;
;           - Ver 0.2.1 (21/04/2006): Eliminados warnings de CMUCL
;             (ver "(declare (ignore ..."). Aadida documentacin.
;
;           - Ver 0.3 (29/04/2006): Empaquetado, puesto interface
;             de definicin de agentes (const agentes). Ofuscado
;             cdigo de agente-humano.
;
;           - Ver 0.4 (03/05/2006): Eliminado bug que impeda jugar
;             un agente humano solo.
;
;           - Ver 0.5 (03/05/2006):  Cambio en agente-humano-fin
;             debido a cambio en tron-comm de ver "fun-0.2".
;
;           - Ver 0.5.1 (05/05/2006): Inclusin de warning si se
;             carga con GCL. 
;
;
; Por hacer:
;           - Correccin de Bugs
;             ------------------
;           - Problema con agente humano gcl que necesita NL
;             despes de pulsar las teclas de movimiento
;
;
;           - Mejoras en funcionalidades
;             --------------------------
;           - Cambiar teclas jugado humano 1?
;           - Prepararlo para otras variantes de lisp (Windows?).
;           - Agregar agente-humano-asistido, slo acepte movimientos
;             vlidos.
;
;           - Mejoras en estilo de programacin
;             ---------------------------------
;
;


;
; Definicin del package
; ----------------------

(defpackage "TRON-AGENTES-HUMANOS"
  (:nicknames "TR-AGENTES-HUMANOS" "TRON-A-H")
  (:use "LISP" "TRON-COMM")
  (:export "AGENTES" "VERSION-TRON-AGENTES-HUMANOS" "AGENTE-HUMANO" "AGENTE-HUMANO-INI" "AGENTE-HUMANO-FIN"))

(in-package "TRON-AGENTES-HUMANOS")


;
; Constante con lista de agente presentes en este package
; -------------------------------------------------------

(defvar agentes '((agente-humano agente-humano-ini agente-humano-fin))
  "Agente humano, efectua los movimientos indicados por teclado.")


;
; Constante versin
; -----------------

(defconstant *version-tron-agentes-humanos* "0.5"
  "Versin del package tron-agentes-humanos.")


;
; Variables globales
; ------------------

(defvar echo-hang t
  "Variable que indica estado echo-hang de la salida/entrada.")

(defvar fun-refresco
  (or (find-symbol "ENVIO-REFRESCO-TAB" "TRON-VIS")
      #'null)
  "Nombre de la funcin para refrescar la pantalla.")


;
; Funciones especficas segn implementacin lisp
; -----------------------------------------------

(let ((lit (lisp-implementation-type)))
  (cond
   ((string-equal lit "GNU Common Lisp (GCL)")
    ; GNU gcl
    ; -------
    (my-warning "En esta versin ~A existe un problema con la lectura de pulsaciones de teclas para el agente humano, no son leidas hasta que no se pulse la tecla enter." lit)
    (load (concatenate 'string
		     (directory-namestring
		      (eval (or (find-symbol "*LOAD-TRUENAME*")
				(find-symbol "*LOAD-PATHNAME*")
				(find-symbol "*LOAD-PATHNAME*" "SYSTEM")
			       "")))
		     "echo-hang-off-gcl")))

  ((string-equal lit "CMU Common Lisp")
    ; CMUCL
    ; -----
   (load (concatenate 'string
		     (directory-namestring
		      (eval (or (find-symbol "*LOAD-TRUENAME*")
				(find-symbol "*LOAD-PATHNAME*")
				(find-symbol "*LOAD-PATHNAME*" "SYSTEM")
			       "")))
		     "echo-hang-off-cmu")))

   ((string-equal lit "CLISP")
     ; GNU clisp
     ; ---------
    (load (concatenate 'string
		     (directory-namestring
		      (eval (or (find-symbol "*LOAD-TRUENAME*")
				(find-symbol "*LOAD-PATHNAME*")
				(find-symbol "*LOAD-PATHNAME*" "SYSTEM")
			       "")))
		     "echo-hang-off-clisp")))

   (t
     ; Desconocida
     ; -----------
    (format *error-output* "Implementacin ~S no soportada" lit)

    (defun echo-hang-off ()
      (setf echo-hang nil))

    (defun restore-echo-hang ()
      (setf echo-hang t))

    (defun read-char-no-hang-no-echo (&optional (stream *standard-input*))
      (read-char-no-hang stream))

    (defun unread-char-no-hang-no-echo (char &optional (stream *standard-input*))
      (unread-char char stream)))))


;
; Agente humano bsico
; --------------------

(defconstant tecla-mov-1
  ; Jugador 1: mov. vi
  '((#\k 0) (#\K 0) (#\l 1) (#\L 1) (#\j 2) (#\J 2) (#\h 3) (#\H 3))
  "Lista asociativa con las teclas y movimientos del jugador 1.")

(defconstant tecla-mov-def
  ; Jugador 2 o por defecto: mov. cursores
  '((#\8 0) (#\A 0) (#\6 1) (#\C 1) (#\2 2) (#\B 2) (#\4 3) (#\D 3))
  "Lista asociativa con las teclas y movimientos del jugador por defecto o jugador 2 si juegan 2 humanos.")

(defvar tecla-mov-jugs nil
  "Lista asociativa con jugadores y lista asociativa de las teclas y movimientos del cada uno.")


(defun agente-humano-ini (jug tipo-juego ags-jug)
  "Funcin que inicializa el agente-humano."
  (declare (ignore jug tipo-juego ags-jug))
  (if echo-hang
      (progn
	(setf tecla-mov-jugs (list (cons jug tecla-mov-def)))
	; Ponemos echo a off, evitar bloquear lecturas y
        ; leer inmediatamente
	(echo-hang-off))
    ; Para dos jugadores humanos (no comer carcteres de oponente)
    (setf tecla-mov-jugs (list (cons 1 tecla-mov-1) (cons 2 tecla-mov-def))))
  (clear-input)
  t)


(defun agente-humano-fin (jug resultado-juego)
  "Funcin que cierra la ejecucin del agente-humano."
  (declare (ignore jug resultado-juego))
  (if (not echo-hang)
      (progn
	; Restauramos estado tty y limpiamos buffer
	(restore-echo-hang)
	(sleep 2)
	(clear-input)))
  t)


(defun agente-humano (jugador)
  "Agente humano que lee los movimientos del jugador Tron JUGADOR (1  2) desde teclado. Solo utiliza las teclas del cursor (8, 6, 2, 4), cuando se llama con dos agentes humanos utiliza para el jugador 1 (izquierdo) las teclas de movimientos del vi(1) (k, l, j, h), y para el jugador 2 (derecho) las teclas del cursor. La barra espaciador refresca la pantalla, siempre que se tenga cargado el agente de visualizacin del package tron-vis."
  (let* ((car (read-char-no-hang-no-echo))
	 (mov (first (rest (assoc car (rest (assoc jugador tecla-mov-jugs)))))))
    (cond (mov mov)
	  ; Refresco pantalla
	  ((eq car #\Space)
	   (funcall fun-refresco) nil)
	  ; Devolvemos car si no era para nosotros, far play
	  ((assoc car (rest (assoc (- 3 jugador) tecla-mov-jugs)))
	   (unread-char-no-hang-no-echo car) nil))))
