; Programa: Package para entorno base para agentes de juego Tron,
;           juego tambin denominado "light cycle",
;           ver http://en.wikipedia.org/wiki/Light_cycle
;
; Asignatura: Introduccin a la Inteligencia Artificial
;             ETSII - ULL
; Prctica: (sistemas de produccin)
; Profesor: Patricio Garca (patricio AT etsii.ull.es)
; Fecha:    07/04/2006
; Comentarios:
;           Este package es el encargado de controlar el desarrollo
;           de una partida Tron entre agentes jugadores de Tron.
;           Almacena todo el estado del juego (tipo de partida, 
;           tablero, posicin de jugadores, direccin de movimiento
;           de stos, ...).
;
;           Se comunica con los agentes mediante un package
;           que hace de interface de comunicaciones denominado
;           tron-comm. De esta forma tron-base se mantiene
;           independiente del sistema de comunicacin empleado
;           (llamadas a funcin, mensajes va sockets, ...). 
;
;           Los tipos de partidas que puede controlar se definen
;           mediante el parmetro TIPO pasado a la funcin
;           INI-TRON, (ver ). ste almacena los siguientes datos:
;
;              - Nmero de jugadores (agentes)
;              - Dimensiones del tablero (x y)
;              - Sin-bordes o con bordes exteriores (t  nil)
;              - Centsimas seg. de retardo entre movimientos
;              - No-verbose o verbose, indica la info. emitida
;              - Juego sncrono (nil) o asncrono (t), esto es,
;                si se espera o no por movim. de agentes (hang)
;              - Controlar el tiempo entre movimientos en servidor
;                de juego (nil) o no (t) (para ejecuciones masivas).
;                No se transmite a los agentes est ltimo.
;
;           Para representar el tablero se utilizan estructuras
;           tipo TRON (ver (describe 'TRON)).
;
;           Las funciones accesibles ms importantes, incluidas en
;           "tron-base-package.lsp", permiten recibir mensajes
;           originarios en los agentes mediante la interface
;           de comunicacin (tron-comm).
;           Ver la documentacin existente para cada una de ellas.
;
;
; Dependencias:
;           Hace uso de funciones del paquete tron-comm (buscar 
;           "tron-comm:") que lleva la comunicacin del juego.
;
; Modificaciones:
;
;           - Ver 0.0 (07/04/2006): Comienzo de la codificacin
;
;           - Ver 0.1 (11/04/2006): Versin inicial con cdigo utlizable
;             para gcl, aunque hay cdigo para CMUCL y opciones para 
;             insertar clisp. Realizado generador de estructura de
;             problemas, funcciones de acceso/modificacin, interface
;             funcional bsica con agentes, visualizador de partidas y
;             agentes humano, rombo, espiral y aleatorio.
;
;           - Ver 0.2 (13/04/2006): Divisin en dos paquetes: "tron-base"
;             (tron-base.lsp) y "trom-comm" (tron-comm-fun.lsp)
;             y diversos agentes (tron-vis-xterm.lsp, 
;             tron-agentes-humanos.lsp y tron-agentes-simples.lsp)
;             Uso de cargadores de paquetes tron-base-package.lsp y
;             tron-comm-fun-package.lsp.
;             Codificado para jugar sin bordes (salto dimensional).
;
;           - Ver 0.2.1 (29/04/2006): Reparado bug "Error en jugar
;             sin bordes, no sale bien por arriba ni por la izquierda.
;             Cheque de posiciones fuera devuelven obstculo."
;
;           - Ver 0.2.2 (11/05/2006): Mejorada documentacin.
;
;           - Ver 0.2.3 (30/05/2006): Aadido :sleep-nil en tipo-juego.
;
;
; Por hacer:
;           - Correccin de Bugs
;             ------------------
;           - Chekeos de consistencia de datos de entrada
;           - Verificar funcionamiento de todo compilado
;
;
;           - Mejoras en funcionalidades
;             --------------------------
;           - Probar todo bajo otros lisps
;             (gcl, clisp linux clisp win, cmucl, ...)
;           - Verificar posiblidad de jugar con /= 2 jugadores
;           - Preparar modo de instalacin
;           - simplificara que jugador 1 = 0 y jugardor 2 = 1?
;           - Distintos tipos de verbose: sin vis, vis pobre,
;             media, rica
;           - Ver que poner con verbose y que con visualizador
;             de partida
;           - Diferenciar verbose (logs) de tipo de visualizador
;           - Calcular tiempos que tardan en decidir los agentes
;           - Buscar implementacin alternativa a contador
;             (con eventos?)
;           - Identificador de partida que se transmita siempre?
;           - ojo, el ltimo movimiento no se informa ok?
;           - Un query sobre el estado del juego?
;           - (setf *tablero-tron* nil) al final de la partida?
;           - Campos de tron-pos-jug y dir-jug como matrices?
;           - Modificar para admitir inicializaciones de ms de 
;             8 jugadores
;           - Hacer el cdigo reentrante
;           - Permitir ejecutar de golpe varias partidas
;           - Juego variante Tron con tamao de cola mximo,
;             nueva tag :cola en tipo-juego.
;
;
;           - Mejoras en estilo de programacin
;             ---------------------------------
;           - Estructurar disposicin de cdigo
;           - Revisar nombre de variables, sobre todo exportadas
;           - Mejorar la documentacin de cada funcin importante
;           - Reescribir ms claramente algunas partes del cdigo...
;           - Cambiar nmeros por smbolos en: direcciones, ...
;           - llamadas con tron-comm: es lo mejor?
;           - Accesos impedidos a tron-mat tron-....? 
;           - Vale la pena la estructura de tron-dir-jug
;             (ver check-set-inc-pos)?
;           - Meter en estructura tron tipo-juego?
;           - Corregir problemas de valores por defecto
;             no asignados a varialbes en parmetros de entrada
;
;
;

(unless (find-package "TRON-BASE")
  (load "tron-base-package"))

(in-package "TRON-BASE")


; Variables version
; -----------------

(defconstant *version-tron-base* "0.2.3" "Versin del package tron-base.")

(defun version-tron-base ()
  "Devuelve la versin del paquete tron-base que se ha cargado."
  *version-tron-base*)


; Constantes
; ----------

(defconstant *obs-tron* 0
  " - Constante: TIPO nmero 

Contiene el nmero que representa los obstculos (#) dentro de la matiz del campo MAT de la estructura tipo TRON.")


; Estructura de representacin del problema a resolver
; ----------------------------------------------------

(defstruct tron
  " - Estructura: TRON (DEFSTRUCT MAT ROBOT DEST)
     Package: TRON-BASE

Estructura que almacena un laberinto Tron, junto a las posiciones del robot, los bloque y los destinos. Consta de tres campos: MAT es una matriz de dos dimensiones, ROBOT y DEST son listas.

La matriz MAT representa el laberinto y los bloque a mover, y tiene por lo tanto las dimensiones mximas de ste. Las paredes del laberinto (#) se representa con el smbolo contenido en la variable *OBS-TRON*, los bloques ($) se representan con el smbolo contenido en la variable *BLQ-TRON* (ver (help '*OBS-TRON*) y (help '*BLQ-TRON*))'. Los espacios vacios estan representados mediante NIL.

El campo ROBOT almacena una lista con dos nmeros, indicando la fila y columna donde esta el robot (@) actualmente.

El campo DEST guarda la informacin de los destinos (.) donde se deberan colocar los bloques, consta de una lista con sublistas de dos elemento indicando fila y columna del destino, contendr tantas sublistas como bloques existan en MAT.

La mayora de funciones de este paquete hace uso de este tipo de estructura."
  mat
  pos-jug
  dir-jug)


; Variables Globales
; ------------------

(defvar *tablero-tron* nil
  "Guarda le estructura tipo TRON del juego actual")

(defvar *dim-tab-tron* nil
  "Dimensiones del la matriz MAT del campo TRON (tablero)")

(defvar *sin-bordes-tron* t
  "Indica si el tablero tiene obstculos en los bordes (NIL) o no (T)")

(defvar *num-jug-tron* 2
  "Numero de jugadores en la partida actual")

(defvar *asincrono* nil
  "Indica si la partida actual es asncrona (NIL) o sncrona (T)")

(defvar *end-tron* (1+ *num-jug-tron*)
  " - Variable: TIPO nmero 

Contiene el nmero que representa la posicin de los choques (!) dentro de la matiz del campo MAT de la estructura tipo TRON.")


; Funciones para el acceso  a la estructura tron
; ----------------------------------------------

(defmacro pos-tab (tab listxy)
  "Posicin x y del tablero"
  `(aref (tron-mat ,tab) (first ,listxy) (second ,listxy)))

(defmacro pos-jug (tab jug)
  "Devuelve la posicin (x y) del jugador <jug> dentro del tablero <tab>"
  `(nth (1- ,jug) (tron-pos-jug ,tab)))

(defmacro dir-jug (tab jug)
  "Direccin de movimiento del jugador jug (^pos [1-1])"
  `(nth (1- ,jug) (tron-dir-jug ,tab)))

(defmacro pos-inc-jug (tab jug)
  "Direccin a incrementar del jugador jug (^pos)"
  `(first (first (dir-jug ,tab ,jug))))

(defmacro val-inc-jug (tab jug)
  "Suma a incrementar del jugador jug [1  -1]"
  `(second (dir-jug ,tab ,jug)))

(defmacro dir-dir-jug (tab jug)
  "Direccin numrica del jugador jug [1  -1]"
  `(third (dir-jug ,tab ,jug)))

(defun borde-out-tab (posxy)
  "Posicin est en un borde del tablero o fuera de l, devuelve nil o res-out"
  (or (>= 0 (first posxy))
      (>= 0 (second posxy))
      (<= (1- (first *dim-tab-tron*)) (first posxy))
      (<= (1- (second *dim-tab-tron*)) (second posxy))))


; Intrucciones del juego
; ----------------------

(defconstant dir-inc
  (make-array '(4 2) :initial-contents
	      '((1 -1)    ; Norte
		(0 +1)    ; Este
		(1 +1)    ; Sur
		(0 -1)))  ; Oeste
  "Array para transformar de direccin a campo dir-jug de tron")

(defmacro dir2dim-inc (dir)
  `(aref dir-inc ,dir 0))

(defmacro dir2pos-inc (tab jug dir)
  `(nthcdr (dir2dim-inc ,dir) (pos-jug ,tab ,jug)))

(defmacro dir2val-inc (dir)
  `(aref dir-inc ,dir 1))

(defmacro jug2dim-inc (tab jug)
  `(nth (aref dir-inc (dir-dir-jug ,tab ,jug) 0) *dim-tab-tron*))

(defun change-dir (tab jug dir)
  "Cambia la direccin de un jugador y devuelve la nueva direccin DIR o NIL si la direccin era la misma"
  (let ((dirj (dir-jug tab jug)))
    (if (and (member dir '(0 1 2 3)) (not (eq (third dirj) dir)))
	(progn
	  (setf (first dirj) (dir2pos-inc tab jug dir))
	  (setf (second  dirj) (dir2val-inc dir))
	  (setf (third dirj) dir)))))

(defun test-mark-mat-pos (tab posxy mark-nil &optional mark-no-nil)
    "Marcamos en la posicin posxy de matriz, si est a nil, con mark, si no con mark-no-nil y devolvemos nil o t respectivamente"
    (if (pos-tab tab posxy)
	(progn
	  (if mark-no-nil
	      (setf (pos-tab tab posxy) mark-no-nil))
	  t)
      (progn (setf (pos-tab tab posxy) mark-nil)
	     nil)))

(defun manten-dentro (val lon-dim)
  "Si se sale del tablero volver por otro lado."
  (cond
   ((>= val lon-dim) (rem val lon-dim))
   ((< val 0) (rem (+ lon-dim val) lon-dim))
   (t val)))

(defun check-set-inc-pos (tab)
  "Incrementa las posiciones de los jugadores y chequea si hay conflicto, devuelve nil  1, 2  3 segn conflicto (3 = ambos)"
  ; Avanzamos los jugadores
  (dotimes (i *num-jug-tron*)
    (let ((jug (1+ i)))
      (setf (pos-inc-jug tab jug)
	    (manten-dentro (+ (pos-inc-jug tab jug) (val-inc-jug tab jug))
			   (jug2dim-inc tab jug)))))
  ; Chequemos conflictos
  (do* (res
	(jug 1 (1+ jug))
	(list-pos-jug (tron-pos-jug tab) (rest list-pos-jug))
	(posxy (first list-pos-jug) (first list-pos-jug)))
      ((endp list-pos-jug) (reverse res))
    (if (member posxy (rest list-pos-jug) :test #'equal)
	; Choca con otro, marcar con *end-tron* y aade a lista de chocados
	(progn
	  (test-mark-mat-pos tab posxy *end-tron* *end-tron*)
	  (push jug res))
      ; No choca con otro
	   ; Mira si choca resto y marca
      (if (test-mark-mat-pos tab posxy jug *end-tron*)
	  (push jug res)))))


; Problema inicial e interno
; --------------------------

(defun jug1-pos-ini ()
    "Posicin inicial del jugador 1"
  (list (floor (1- (first *dim-tab-tron*)) 4) (floor (1- (second *dim-tab-tron*)) 2)))

(defun jug2-pos-ini ()
    "Posicin inicial del jugador 2"
  (list (ceiling (* (1- (first *dim-tab-tron*)) 3) 4) (floor (1- (second *dim-tab-tron*)) 2)))

(defun jugs-pos-ini (n)
  "Posiciones iniciales de los N jugadores"
  ; Ojo, solo admite 2 variantes (8 jugadores)
  (let (pos-ini)
    (dotimes (i (ceiling n 2) (reverse (nthcdr (rem n 2) pos-ini)))
      (push (jug1-pos-ini) pos-ini)
      (push (jug2-pos-ini) pos-ini))))

(defconstant jug1-dir-ini 1   ; Este
  "Direccin inicial del jugador 1 en el tablero por defecto")

(defconstant jug2-dir-ini 3   ; Oeste
  "Direccin inicial del jugador 2 en el tablero por defecto")

(defun jugs-dir-ini (n)
  "Direcciones iniciales de los N jugadores"
  ; Ojo, solo admite 4 variantes (8 jugadores)
 (let (dir-ini
       (cont1 jug1-dir-ini)
       (cont2 jug2-dir-ini))
    (dotimes (i (ceiling n 2) (reverse dir-ini))
      (push cont1 dir-ini)
      (push cont2 dir-ini)
      (setf cont1 (rem (1- cont1) 4))
      (setf cont2 (rem (1+ cont2) 4)))))

(defun mark-borde-tab (tab dim-tab)
  "Marcar los bordes del tablero"
  (dotimes (x (first dim-tab))
    (test-mark-mat-pos tab (list x 0) *obs-tron*)
    (test-mark-mat-pos tab (list x (1- (second dim-tab))) *obs-tron*))
  (dotimes (y (second dim-tab))
    (test-mark-mat-pos tab (list 0 y) *obs-tron*)
    (test-mark-mat-pos tab (list (1- (first dim-tab)) y) *obs-tron*)))


; Generacion estructura inicial del problema a resolver
; -----------------------------------------------------

(defun gen-tron-def (num-jug &optional dim-tab sin-bordes)
  "Genera tablero por defecto de jugadores y dimensiones pasados"
  (if dim-tab
      (setf *dim-tab-tron* dim-tab)
    (setf *dim-tab-tron* dim-tab-def))
  (setf *sin-bordes-tron* sin-bordes)
  (setf *num-jug-tron* num-jug)
  (let ((tab-ini (make-tron)))
    ; Inicializamos los campos de la estructura tab-ini
    (setf (tron-mat tab-ini) (make-array *dim-tab-tron* :initial-element nil)
	  (tron-pos-jug tab-ini) (jugs-pos-ini num-jug)
	  (tron-dir-jug tab-ini) nil)
    ; Colocamos direccion de movimiento inicial
    (dotimes (i *num-jug-tron*)
      (push (list nil 0 0) (tron-dir-jug tab-ini)))
    (let ((dirs (jugs-dir-ini num-jug)))
      (dotimes (i *num-jug-tron*)
	(change-dir tab-ini (1+ i) (nth i dirs))))
    ; Marcamos en la matriz las posiciones iniciales de jugadores
    (dotimes (i *num-jug-tron*)
      (test-mark-mat-pos tab-ini (pos-jug tab-ini (1+ i)) (1+ i)))
    (if (not *sin-bordes-tron*)
        ; Marcamos en la matriz los bordes del tablero
	(mark-borde-tab tab-ini *dim-tab-tron*))
    ; Devolvemos tablero
    tab-ini))

 
(defun gen-tron (&optional (num-jugadores 2) dim-tab sin-bordes)
  " - Funcin: GEN-TRON (NUM-JUGADORES DIM-TAB SIN-BORDES)
     Package: TRON-BASE

Genera un estado inicial con un laberinto para el Tron, de acuerdo a las especificiaciones del argumento TIPO pasado a la funcin INIC-TRON, puede cargar el laberinto desde un fichero si asi se le ha especificado en esta variable (ver (help 'INIC-TRON). Devuelve una estructura tipo TRON."
  (gen-tron-def num-jugadores dim-tab sin-bordes))


;
; Interface de consultas a servidor de juego
; ------------------------------------------


(defun pos-jug-tron (jugador)
  "Consulta posicin actual del jugador con identificador JUGADOR. El resultado, si JUGADOR corresponde al identificador de un jugador vlido, es una lista con dos nmeros correspondientes al valor en las coordenadas X e Y. Las coordenadas comienzan (0 0) en la esquina superior izquierda. Si JUGADOR es un identificador de un jugador no vlido se devuelve NIL."
  (copy-list (pos-jug *tablero-tron* jugador)))

(defun dir-jug-tron (jugador)
  "Consulta direccin actual del jugador con identificador JUGADOR. El resultado, si JUGADOR corresponde al identificador de un jugador vlido, es un nmero entre 0 y 4, correspondientes a las direcciones de Norte, Este, Sur y Oeste respectivamente. Si JUGADOR es un identificador de un jugador no vlido se devuelve NIL."
  (dir-dir-jug *tablero-tron* jugador))

(defun pos-inc-dir (pos dir)
  "Calcula la posicin resultante de desplazarse en la direccinb DIR respecto a POS."
  (let ((pos-inc (copy-list pos)))
    (setf (nth (dir2dim-inc dir) pos-inc)
	  (+ (nth (dir2dim-inc dir) pos-inc)
	     (dir2val-inc dir)))
    pos-inc))

(defun pos-tab-tron (pos &optional dir)
  "Contenido actual de una posicin del tablero POS ms opcionalmente desplazamiento en una direccin DIR sobre ella. Los valores de POS y DIR han de ser como se indican en POS-JUG-TRON y en DIR-JUG-TRON. El resultado devuelto es NIL si la casilla est vaca. Si la casilla est ocupada se devuelve un nmero que es igual al identificador del jugador que marc dicha casilla, o a 0 en caso de que contenga un obstculo (borde del tablero)."
  ; Chequeo de parmetros
  (if (and (consp pos) (numberp (first pos)) (numberp (second pos))
	   (or (null dir) 
	       (and (numberp dir) (>= dir 0) (<= dir 3))))
      ; Clculo de posicin a chekear
      (let ((pos-to-check 
	     (if (not dir)
		 pos
	       (pos-inc-dir pos dir))))
        ; Chequeo de bordes y resultado
	(if (borde-out-tab pos-to-check)
	    (if *sin-bordes-tron*
		(progn
		  (setf (first pos-to-check)
			(manten-dentro (first pos-to-check)
				       (first *dim-tab-tron*))
			(second pos-to-check)
			(manten-dentro (second pos-to-check)
				       (second *dim-tab-tron*)))
		  (pos-tab *tablero-tron* pos-to-check))
		*obs-tron*)
	  (pos-tab *tablero-tron* pos-to-check)))))

(defun copy-array (arr)
  "Genera una copia de un array de dos dimensiones."
  (let ((arr2 (make-array `(,(array-dimension arr 0)
		,(array-dimension arr 1)))))
    (dotimes (i (array-dimension arr 0) arr2)
      (dotimes (j (array-dimension arr 1))
	(setf (aref arr2 i j) (aref arr i j))))))

(defun tab-tron ()
  "Solicita la matriz del tablero Tron, devolviendolo en un array."
  (copy-array (tron-mat *tablero-tron*)))


; Contador de tiempo entre movimientos
; ------------------------------------

(defvar counter nil
  "Vairable contador.")

(defun ini-counter (limit)
  "Inicializa el contador a LIMIT centsimas de segundo."
  (setf counter
	(if limit
	    (+ (get-internal-real-time)
	       (/ (* limit internal-time-units-per-second)
		  100)))))

(defun test-counter (&optional hang)
  "Chequea si el contador llegado al lmite, esperando si HANG es T."
  (if counter
      (do ()
	  ((> (get-internal-real-time) counter) t)
	(if (not hang)
	    (return nil)))
    hang))


; Obtencin de movimientos
; ------------------------

(defvar lis-mov nil
  "Lista de movimientos a efectuar")

(defvar lis-jug nil
  "Lista de jugadores")

(defun get-accion (jug dir)
  (push (list jug dir) lis-mov))

(defun obt-lis-mov ()
  (if *asincrono*
      (prog1
	lis-mov
	(setf lis-mov nil))
      (remove-if-not #'second 
		     (mapcar #'(lambda (jug)
				 (list jug
				       (tron-comm:consulta-accion jug)))
			     lis-jug))))

; Lanzador de agentes y problemas
; -------------------------------

(defvar num-jug-def 2
  "Nmero de jugadores por defecto")

(defvar dim-tab-def '(80 23)
  "Dimensiones del tablero por defecto (x y)")

(defvar sleep-cs-def 15
  "Centsimas de segundo entre movimientos por defecto")

(defun extrc-tipo (tipo)
  "Extraccin del tipo de partida a efectuar"
  (values
   (or (first tipo) num-jug-def)     ; Nmero de jugadores
   (or (second tipo) dim-tab-def)    ; Dimensiones del tablero (x y)
   (third tipo)                      ; sin-bordes (t  nil)
   (or (fourth tipo) sleep-cs-def)   ; cs de retardo (nmero)
   (fifth tipo)                      ; no-verbose
   (sixth tipo)                      ; sncrono (nil) o asncrono (t)
   (seventh tipo)))                  ; controlar (nil) o no (t) tiempo mov.

(defun ini-tron (&optional tipo)
  "Lanza una partida TRON del tipo TIPO (ver EXTRC-TIPO)"
  ; Establecemos valores por defecto del tipo de juego
  (multiple-value-bind (num-jug dim-tab sin-bordes sleep-cs no-verbose asincrono sleep-nil)
      (extrc-tipo tipo)
    (setf *asincrono* asincrono)
    ; Generamos tablero de juego
    (setf *tablero-tron* (gen-tron num-jug dim-tab sin-bordes))
    (setf *end-tron* (1+ num-jug))
    ; Inicializamos contador
    (ini-counter (and (not sleep-nil) sleep-cs))
    ; Inicializamos lista de jugadores
    (setf lis-jug nil)
    (dotimes (i num-jug)
      (push (1+ i) lis-jug))
    ; Inicializamos los agentes
    (tron-comm:envio-ini (list num-jug
			       dim-tab
			       sin-bordes
			       sleep-cs
			       no-verbose
			       asincrono))

    ; Inicio bucle iteraciones ---------------------------------------------
    (do* (          ; Obtencin 'concurrente' de movimientos
	  (lis-movs nil 
		    (obt-lis-mov))
	                 ; Generacin de cambios de direccin
	  (lis-real-movs nil 
			 (remove-if-not #'second
					(mapcar #'(lambda (mov)
						    (list mov
							  (change-dir *tablero-tron*
								      (first mov)
								      (second mov))))
						lis-movs)))
                   ; Incrementa posiciones y chequea fin
	  (res nil (check-set-inc-pos *tablero-tron*)))
        ; Condicin de salida ----------------------------------------------
	(res
	 ; Indicamos a los agentes fin de patrida,
         ; ojo, el ltimo movimiento no se informa
	 (tron-comm:envio-fin res)
	 res)

      ; Cuerpo del bucle -------------------------------------------------
      ; Informa si ha habido cambio de direccin
      (mapcar #'(lambda (mov)
		  (tron-comm:envio-camb-est
		   (first mov)
		   (second mov)))
	      lis-real-movs)
      (if (not no-verbose)
          ; Informa de cambios de posicin
	  (mapcar #'(lambda (jug)
		      (tron-comm:envio-camb-pos
		       jug
		       (pos-jug *tablero-tron* jug)))
		  lis-jug))
      ; Esperamos por el contador
      (test-counter t)
      (ini-counter (and (not sleep-nil) sleep-cs)))))
      ; Fin del bucle de iteraciones --------------------------------------
