; Programa: Package para entorno interface funcional de
;           comunicaciones con agentes de juegos,
;           utilizado para el juego Tron.
;
; Asignatura: Introduccin a la Inteligencia Artificial
;             ETSII - ULL
; Prctica: (sistemas de produccin)
; Profesor: Patricio Garca (patricio AT etsii.ull.es)
; Fecha:    12/04/2006
; Comentarios:
;           Este package es el encargado de comunicar a los agentes
;           jugadores, visualizadores y/o logs con el servidor
;           o controlador de partidas tron-base y/o con otros
;           agentes.
;
;           Esta implementacin utiliza para dicha comunicacion
;           llamadas a funciones, esto es, emite informacin
;           llamando a funciones exportadas en el servidor del
;           juego(tron-base) y en los agentes, y recibe informacin
;           mediante llamadas a sus funciones que exporta.
;           El hacerlo  de esta forma limita a que el servidor
;           del juego, este package y todos los agentes estn
;           ejecutandose en la msma mquina y en la misma
;           instaciacin del LISP.
;
;           Esta interface de comunicacin aisla tambin al
;           servidor del juego de los agentes, de forma que
;           el servidor no tenga por qu saber que agentes
;           estn jugando (nombres, caractersticas, funciones,
;           descripciones, ...), ni las comunicaciones que existan
;           entre los agentes, etc.
;
;           Bsicamente un agente que participe en una partida
;           puede ser de dos tipos:
;
;              - Agente activo: Normalmente jugadores, se les
;                permite o se les inquiere a efecuar acciones
;                que puedan cambiar el estado del juego
;
;              - Agentes pasivos: Son agentes que observan la
;                partida, aunque opcionalmente pueden atender
;                alguna peticin de otros agentes o el servidor.
;                Entre ellos estaran agentes para la 
;                visualizacin de la partida, agentes para
;                almacenar un log de la partida y/o recabar
;                estadsticas o agentes organizadores de
;                partidas.
;
;           Un agente se compone de un grupo de funciones, que
;           habrn de pasarse a la funcin envio-pet-ini cuando
;           comience una partida. Este grupo de  fuciones pueden
;           ser algunas de las siguientes:
;
;               - Funcin de generacin de accin: Slo
;                 permitidas a agentes activos. Son llamadas
;                 cuando el servidor del juego requiera una
;                 accin del agente, pasndole un identificador
;                 del jugador. El agente activo ha de devolver
;                 la accin solicitada.
;
;               - Funcin de inicio de juego: Se llama a esta
;                 funcin del agente cuando se inicia una
;                 partida, pasndosele el identificador de
;                 jugador, si es agente activo, o nil si es
;                 agente pasivo; tipo de juego de la partida
;                 que se inicia y lista de jugadores activos.
;
;               - Funcin de fin de juego: Se llama cuando
;                 el juego ha finalizado, pasndosele el 
;                 identificador de jugador, si es agente activo,
;                 o nil si es agente pasivo, y el resultado final
;                 del juego.
;
;               - Funcin de seguimiento de posicin de
;                 jugadores: Cada vez que un jugador se desplaza
;                 se llama a esta funcin, pasndole el el 
;                 identificador del jugador que se ha desplazado
;                 y su nueva posicin.
;
;               - Funcin de seguimiento de acciones de
;                 jugadores: Cada vez que un jugador efectua una
;                 accin que modifica el estado del juego se llama
;                 a esta funcin, pasndole el identificador del 
;                 jugador que ha efectuado la accin y su nuevo
;                 estado.
;
;               - Adems, en el futuro podrn aadirse:
;                    - funcin-seguimiento-consultas-contenido-posicion:
;                    - funcin-seguimiento-consultas-posicion-jugador
;                    - funcin-seguimiento-consultas-estado-jugador
;
;           Los agentes podrn llamar, desde el cdigo de cualquiera
;           de las anteriores funciones a las siguientes funciones:
;
;                    - ENVIO-PET-INI: Solicitud inicio partida
;                    - ENVIO-PET-CONSULTA-ACCION: Solicitud de que
;                      se le consulte una accin
;                    - CONSULTA-POS-JUG: Consultar posicin de jugador
;                    - CONSULTA-EST-JUG: Consultar estado de jugador
;                    - CONSULTA-POS-TAB: Consultar posicin de tablero
;                    - CONSULTA-TAB: Recabar todo el tablero
;
;           Estas y otras funciones accesibles importantes son las
;           incluidas en "tron-comm-fun-package.lsp".
;           Ver la documentacin existente para cada una de ellas.
;
;           Se intenta que este package sea independiente del
;           juego que sirva el controlador del juego. Por ello
;           las referencias/dependencias al juego Tron se
;           iran tratando de eliminar.
;
;           Para utilizar una interface no funcional ha de
;           reescribirse el cdigo, cambiando bsicamente las
;           primitivas funcall y apply por otras equivalentes
;           al sistema de comunicacin a emplear.
;
;
; Dependencias:
;
;           Hace uso de funciones del paquete tron-base que 
;           lleva el control del juego.
;
;
; Modificaciones:
;
;           - Ver 0.1 (13/04/2006): Primera versin. Extraida
;             de tron.0.1.lsp. Vease tron-base.lsp
;
;           - Ver 0.2 (04/05/2006): Cambio envios a funciones
;             de finalizacin, inclusin de identificador de
;             jugador para agentes activos o nil para agentes
;             pasivos. Inclusin de funciones envio-evento*.
;             En envio-ini se enva listas con sublistas de
;             strings agentes - descripcin de agentes
;             jugadores, y no slo smbolos.
;
;           - Ver 0.2.1 (11/05/2006): Mejorada documentacin.
;             cambio nombre funcin envio-camb-est. Cambio
;             nombre variables lista de funciones.
;
;           - Ver 0.3 (12/05/2006): Incluida funciones para
;             el seguimiento de consultas en envio-pet-ini.
;             Modificaciones de funciones consulta-* para 
;             permitir seguimiento.
;
;           - Ver 0.3.1 (30/05/2006): Incluida funcin 
;             consulta-turno y consulta-contador.
;
;
; Por hacer:
;           - Correccin de Bugs
;             ------------------
;           - Escoger elementos a exportar
;
;
;           - Mejoras en funcionalidades
;             --------------------------
;           - Precisar param-juego y resultado-juego
;           - Pasar a servidor juego lista deresultados de cada
;             llamada dentro de envio-evento*?
;           - Seleeccionar valor a devolver por pet-ini
;           - En vez de empezar con dim-tab-def fija obtener tamao
;             desde la terminal?
;           - Utilizar interface de comunicaciones va sockets o
;             similar (irc, http, kqml, pvm, ...)
;           - Aadir un gestor ms generico (otro paquete?) que
;             permita llamadas ms simples a pet-ini-tron
;           - Intentar independizar de juego
;           - Separar interface para servidor de interface agentes?
;             (seguridad)
;           - Definicin de my-warning est bien aqu?
;           - Hacer que tipo-juego sea lista de tags?:
;             (:dimension (80 25) :asincrono t ...)
;           - Hacer que consultas con parmetro jugador acepte, adems
;             string con nombre de jugador.
;           - Que se les pase a las funciones de seguimientos de
;             consultas el que solicit la consulta.
;
;
;           - Mejoras en estilo de programacin
;             ---------------------------------
;           - Unificar/armonizar interface de llamadas de agentes
;           - Unificar/acordar estructura de cdigo de agentes
;           - Revisar nombre de variables, sobre todo exportadas
;           - Agregar la documentacin a cada funcin importante
;           - Poner la documentacin correcta
;           - utilizar defconstant para constantes?
;           - llamadas con tron-base: es lo mejor? (dependen juego)
;           - Cambiar parmetros de send-ini-tron y
;             llamadas a funciones de ini de agentes
;


(unless (find-package "TRON-COMM")
  (load "tron-comm-fun-package"))

(in-package "TRON-COMM")


;
; Variables versin
; -----------------

(defconstant *version-tron-comm* "fun-0.3.1" "Versin del package tron-comm.")

(defun version-tron-comm ()
  "Devuelve la versin del paquete tron-comm que se ha cargado."
  *version-tron-comm*)


;
; Variables globales
; ------------------

(defvar fun-gen-est nil
  "Lista con las funciones que generan los cambios de estado de cada jugador, esto es, funciones de agentes jugadores ppd.")

(defvar num-jugadores 2
  "Nmero de jugadores de la partida actual, ha de coincidir con la longitud de fun-gen-est.")

(defvar fun-ini nil
  "Lista con las funciones que han de llamarse al comienzo del juego Tron.")

(defvar fun-fin nil
  "Lista con las funciones que han de llamarse al final del juego Tron.")

(defvar fun-seg-pos nil
  "Lista con las funciones que han de llamarse al haber un cambio de posicin de algn jugador tron, si est el verbose activado.")

(defvar fun-seg-est nil
  "Lista con las funciones que han de llamarse al haber un cambio del estado de algn jugador Tron.")

(defvar fun-seg-cposj nil
  "Lista con las funciones que han de llamarse al haber una consulta de la posicin de algn jugador Tron.")

(defvar fun-seg-cestj nil
  "Lista con las funciones que han de llamarse al haber una consulta del estado de algn jugador Tron.")

(defvar fun-seg-cpost nil
  "Lista con las funciones que han de llamarse al haber una consulta del contenido de la posicin de alguna casilla del tablero Tron.")

(defvar turno-jugador nil
  "Indica quien es el jugador que en estos momentos esta en posesin del turno, o NIL si ninguno.")

;
; Warnings
; --------

(defun my-warning (datum &rest arguments)
  (apply #'format 
	 `(,*error-output* ,(format nil "Warning: ~A~%" datum) ,@arguments)))


; Contador de tiempo entre movimientos
; ------------------------------------

(defvar limite-cs nil
  "Lmite en centsimas de segundo para decidir movimiento.")

(defvar counter nil
  "Vairable contador.")

(defun ini-counter ()
  "Inicializa el contador a LIMIT-CS centsimas de segundo."
  (setf counter
	(if (and limite-cs (> limite-cs 0))
	    (+ (get-internal-real-time)
	       (/ (* limite-cs internal-time-units-per-second)
		  100)))))

(defun test-counter (&optional hang)
  "Chequea si el contador llegado al lmite, esperando si HANG es T."
  (if counter
      (do ()
	  ((> (get-internal-real-time) counter) t)
	(if (not hang)
	    (return nil)))
    hang))


;
; Funciones de apoyo para comunicacin
; ------------------------------------

(defun envio-evento (lis-fun-rec &rest lis-datos)
  "Enviar evento y datos LIS-DATOS las funciones en LIS-FUN-REC, indicando les tamben su identificador de jugador."
  (do ((jug 1 (if (and jug (< jug num-jugadores))
		  (1+ jug)))
       (ag-rec lis-fun-rec (rest ag-rec)))
      ((endp ag-rec) t)
    (if (first ag-rec)
	(progn
	  (setf turno-jugador jug)
	  (if (null (apply (first ag-rec) jug lis-datos))
	      (if jug
		  (progn
		    (error "~S no acepta evento." ag-rec)
		    (return nil))
		(my-warning "~S no acepta evento." ag-rec))))))
  (setf turno-jugador nil))

(defun envio-evento-pasivo (lis-fun-rec &rest lis-datos)
  "Enviar evento pasivo (sin indicacin de id de jugador) y datos LIS-DATOS las funciones en LIS-FUN-REC."
  (dolist (ag-rec lis-fun-rec t)
    (apply ag-rec lis-datos)))


;
; Interface de consultas servidor de juego tron-base  ==> agentes
; ---------------------------------------------------------------

(defun envio-ini (tipo-juego)
  "Enviar indicacin de comienzo de partida."
  (setf num-jugadores (first tipo-juego))
  (setf limite-cs (fourth tipo-juego))
  (envio-evento fun-ini
		tipo-juego
		(mapcar (lambda (agente)
			  ; Obtencion de nombres y descripciones
			  (list (symbol-name agente)
				(documentation agente 'function)))
			  fun-gen-est)))

(defun envio-fin (resultado-juego)
  "Enviar indicacin de fin de partida."
  (envio-evento fun-fin resultado-juego))

(defun consulta-accion (jugador)
  "Solicitar accin a agente con identificador JUGADOR."
  (prog2
      (progn
	(setf turno-jugador jugador)
	(ini-counter))
      (funcall (nth (1- jugador) fun-gen-est) jugador)
    (setf turno-jugador nil)))

(defun envio-camb-est (jugador estado)
  "Enviar indicacin de cambio de estado de jugador con identificador JUGADOR."
  (envio-evento-pasivo fun-seg-est jugador estado))

(defun envio-camb-pos (jugador posicion)
  "Enviar indicacin de cambio de posicin de jugador con identificador JUGADOR."
  (envio-evento-pasivo fun-seg-pos jugador posicion))


;
; Interface de consultas agentes ==> servidor de juego tron-base
; --------------------------------------------------------------

(defun envio-pet-ini (lis-fun-gen-est
		      &optional
		      lis-fun-ini
		      lis-fun-fin
		      lis-fun-seg-pos
		      lis-fun-seg-est
		      lis-fun-seg-cposj
		      lis-fun-seg-cestj
		      lis-fun-seg-cpost
		      tipo-juego)
  "Lanza una partida entre los agentes con funciones que generan movimientos en LIS-FUN-GEN-EST. LIS-FUN-INI ha de ser una lista de funciones para inicializar los agentes al comienzo de una partida, en el mismo orden que LIS-FUN-GEN-EST y pudiendo incluir en la cola agentes pasivos. Si algn agente pasivo no tiene funcin de inicializacin se ha de incluir NIL. LIS-FUN-FIN es similar a LIS-FUN-INIpero para funciones de finalizacin de la partida. LIS-FUN-SEG-POS ha de ser una lista de funciones que realizarn el seguimiento de los cambios de posicin de los jugadores. LIS-FUN-SEG-EST ha de ser una lista de funciones que realizarn el seguimiento de los cambios de estado de los jugadores. TIPO-JUEGO define el tipo de juego que se pretende jugar en la partida (ver servidor de juego TRON-BASE)."
  (setf fun-gen-est lis-fun-gen-est
	fun-ini lis-fun-ini
	fun-fin lis-fun-fin
	fun-seg-pos lis-fun-seg-pos
	fun-seg-est lis-fun-seg-est
	fun-seg-cposj lis-fun-seg-cposj
	fun-seg-cestj lis-fun-seg-cestj
	fun-seg-cpost lis-fun-seg-cpost)
  ; Ajustamos nmero de jugadores
  (cond
   ((null tipo-juego) (setf tipo-juego `(,(length fun-gen-est))))
   ((null (first tipo-juego)) (setf (first tipo-juego) (length fun-gen-est))))
  ; Llamada a servidor tron
  (or (tron-base:ini-tron
                     tipo-juego)           ; Tipo de juego
      (error "funcin envio-pet-ini falla")))

(defun envio-pet-consulta-accion (jugador) ; Para tron asncrono
  "Solicita al servidor del juego que le sea enviado una CONSULTA ACCIN. JUGADOR ha de se el identificador del jugador que solicita la consulta."
  (let ((accion (consulta-accion jugador)))
    (if accion
	(tron-base:get-accion jugador accion))))

(defun consulta-pos-jug (jugador)
  "Posicin actual del jugador con identificador JUGADOR. El resultado depende de lo enviado por el servidor de juegos (ver TRON-BASE:POS-JUG-TRON)."
  (if fun-seg-cposj
      (envio-evento-pasivo fun-seg-cposj turno-jugador jugador))
  (tron-base:pos-jug-tron jugador))

(defun consulta-est-jug (jugador)
  "Estado actual del jugador con identificador JUGADOR. El resultado depende de lo enviado por el servidor de juegos (ver TRON-BASE:DIR-JUG-TRON)."
  (if fun-seg-cestj
      (envio-evento-pasivo fun-seg-cestj turno-jugador jugador))
  (tron-base:dir-jug-tron jugador))

(defun consulta-pos-tab (pos &optional dir)
  "Contenido actual de una posicin del tablero POS ms opcionalmente desplazamiento en la direccin DIR sobre ella. El resultado depende de lo enviado por el servidor de juegos (ver TRON-BASE:POS-TAB-TRON)."
  (if fun-seg-cpost
      (envio-evento-pasivo fun-seg-cpost turno-jugador pos dir))
  (tron-base:pos-tab-tron pos dir))

(defun consulta-tab ()
  "Solicita la matriz del tablero del juego."
  (tron-base:tab-tron))


;
; Interface de consultas agentes ==> servidor de comunicaciones
; -------------------------------------------------------------

(defun consulta-turno ()
  "Solicita el identificador que en estos momentos esta en posesin del turno de juego."
  turno-jugador)

(defun consulta-contador (&optional hang)
  "Consulta si se ha cumplido el tiempo para emitir un movimiento, esperando si HANG es T."
  (test-counter hang))

