; Programa: Cargador de entorno de juego Tron,
;           juego tambin denominado "light cycle",
;           ver http://en.wikipedia.org/wiki/Light_cycle
;
; Asignatura: Introduccin a la Inteligencia Artificial
;             ETSII - ULL
; Prctica: (sistemas de produccin)
; Profesor: Patricio Garca (patricio AT etsii.ull.es)
; Fecha:    28/04/2006
; Comentarios:
;           Las funciones aqu implementadas permiten manejar
;           de una manera cmoda los agentes utilizables
;           para el entorno de juego Tron (ver tron-base y
;           tron-comm). Facilitando la carga en memoria de
;           todo el entorno y los agentes, consulta de los 
;           agentes cargados en memoria y sus caractersticas,
;           ejecucin de partidas entre agentes y transferencia
;           de ficheros de agentes al directorio 'arena'.
;
;           Para poder utilizar estas funciones los agentes
;           han de estar definidos bajo las siguientes normas
;           (ver ejemplo "tron-agente-humanos.lsp" y
;           "tron-agentes-simples.lsp"):
;
;           - Se ha de crear un fichero que contenga la
;           definicin de la/s funcion/es de el/los agentes
;           que se deseen crear.
;           - El fichero debe de contener definir un package
;           cuyo nombre ha de conincidir con el nombre del
;           fichero sin extensin. Dicho package ha de usar
;           el entrono de comunicaciones "TRON-COMM" y ha de
;           exportar todas las funciones que definen cada
;           agente.
;           - Adems el package anterior debe exportar una
;           constante denominada 'agente', dicha constante
;           ha de iniciallizarse en el fichero (defconstant)
;           a una lista con una sublista por cada agente
;           que se quiera definir. Dichas sublistas contendrn
;           las funciones que definen cada agente (ver
;           documentacin de variable lis-agentes y
;           documentacin de tron-comm).
;
;           En este cdigo se define dos variables que indican
;           el directorio y subdirectorio donde se encuentran
;           los ficheros necesarios para cargar el entorno Tron
;           y el repositorio de agentes  'arena'. Estas 
;           variables son respectivamente path-tron y
;           path-tron-arena. Estas varialbes, al cargarse, se
;           inicializan al directorio donde est tron-loader y
;           a el subdirectorio "arena" del directorio donde est
;           tron-loader respectivamente (ver documentacin
;           path-tron y path-tron-arena).
;
;           Este entorno carga automticamente tres ficheros
;           con los siguientes agentes:
;           - "tron-vis" con agente-vis: agente pasivo
;           encargado de la visulizacin de partidas
;           - "tron-agentes-humanos" con agente-humano:
;           agente activo para que un humano pueda jugar
;           - "teron-agentes-simples": posee varios agentes
;           activos sencillos
;
;           Las funciones accesibles mas importantes son:
;
;           load-tron
;           load-tron-arena
;           jug-tron
;           tron
;           ls-tron-arena
;           tron-todos
;           version-tron
;           send-tron-arena
;           del-tron-arena
;           *vis-tron*
;           *dim-tab-tron*
;           *sin-bordes-tron*
;           *sleep-cs-tron*
;           *no-verbose-tron*
;           *asincrono-tron*
;
;           Ver la documentacin existente para cada una de ellas.
;
;           A continuacin se muestran algunos ejemplos de llamadas
;           a las fuciones aqu implementadas:
;
;
;           - Consulta de agentes en memoria:
;             (jug-tron)
;
;           - Iniciar partida por defecto, dos agentes humanos:
;             (tron)
;
;           - Iniciar partida entre dos agentes dados:
;             (tron "AGENTE-ESPIRAL" "AGENTE-HORMIGA-BORRACHA")
;
;           - Carga de fichero de agentes "agentes-aluXXX":
;             (load-tron "agentes-aluXXX")
;
;           - Consulta 'full' de agentes en memoria:
;             (jug-tron t)
;
;          - Jugar todos agentes entre s rpidamente
;            y sin visualizacin:
;             (tron-todos :sleep-cs 0 :vis nil)
;
;
; Dependencias:
;           Hace uso del paquete que contiene la interface (tron-comm)
;           con el control del juego (tron-base).
;
;           Hace uso de los paquetes tron-vis, tron-agentes-humanos
;           y tron-agentes-simples.
;
;
; Modificaciones:
;
;           - Ver 0.1 (28/04/2006): Primera versin.
;
;           - Ver 0.2 (02/05/2006): Incluido tron-todos.
;
;           - Ver 0.3 (03/05/2006): Carga automticamente variable
;             path-tron desde *load-truename*. Implementada 
;             lis-fich-load-arena para controlar ficheros cargados
;             del arena.
;
;           - Ver 0.3.1 (03/05/2006): Corregido bug que confunda
;             el identificador del jugador enviado a funciones de
;             inicializacin de agentes (no enviaba nil, agente
;             humano como jugador 2 confunda teclas). Corregido
;             bug para cargar con GCL (*load-patname* y
;             *load-trueanem*).
;
;           - Ver 0.3.2 (04/05/2006): Ampliada key :vis de tron
;             para contemplar paso de string y abrirlo como stream
;             de salida (*tron-vis-stream*) para tron-vis,
;             ver v. 0.2 de tron-vis-xterm.
;
;           - Ver 0.3.3 (05/05/2006): Aadiad funcin version-tron.
;             Ampliada opcin full de jug-tron. Utilizado
;             merge-pathnames para unir directorios y ficheros.
;             Corregido bug de cargar varias veces tron-vis en gcl. 
;
;           - Ver 0.3.4 (11/05/2006): Implementadas funciones 
;             send-tron-arena y del-tron-arena.
;
;           - Ver 0.4 (12/05/2006): Cambios en tron-internal debido
;             a cambio en tron-comm de ver "fun-0.3". Tratamiento
;             a distinto tipo de visualizadores (:vis) y desplazado
;             dispositivo de visualizacin a :dev-vis.
;
;           - Ver 0.4.1 (18/05/2006): Cambiado send-tron-arena y
;             del-tron-arena para llamar a programs externos
;             almacenados en las variables sender-tron-arena y
;             deleter-tron-arena respectivamente.
;
;           - Ver 0.4.2 (20/05/2006): Modificacines sender y
;             del-tron-arena.
;
;           - Ver 0.4.3 (27/05/2006): Modificacin tron-todos para
;             mostrar nmero de partidas jugadas. Aadido funcin
;             stat-tron-arena con actualizazin del "hof.lsp".
;
;           - Ver 0.4.4 (29/05/2006): Aadido nuevos-tron-arena .
;
;           - Ver 0.5 (30/05/2006): Carga paquetes shell, q_search
;             y a_star. Aadido :sleep-nil en tipo-juego.
;
;
; Por hacer:
;           - Correccin de Bugs
;             ------------------
;           - Chequear funcionamiento en clisp:
;             (directory (concatenate 'string path-tron-arena ""))
;             con resultado distinto.
;           - Versiones que no signan los links para clisp y gcl
;             de load-tron.
;           - Chekear conflictos al cargar agentes con igual nombre.
;           - directory (en load-tron) expande los nombres de
;             ficheros aadiendo * al final. Limitarlos con
;             lis-lisp-types.
;           - Hacer que send-tron-arena y del-tron-arena funcionen
;             para gcl.
;           - send-tron-arena  canta /bin/tar: Member names contain `..
;           - del-tron-arena canta "stty: entrada estndar: 
;             Funcin ioctl no apropiada para el dispositivo"
;
;           - Mejoras en funcionalidades
;             --------------------------
;           - Chequear que pasa si no se ha definido varialble
;             con lista de agentes del fichero en load-tron.
;           - Aadir agente log de partidas.
;           - Implementar funciones: stat-tron
;           - Guardar info que relacione agente - fichero.
;           - Funcin que descarge agentes de memoria (unload)?
;             (delete-package).
;
;           - Mejoras en estilo de programacin
;             ---------------------------------
;           - Pasar al campo :documentation de defpackage
;             la documetacin de cabeceras.
;           - Esto como package?
;           - Aadir especificaciones parmetros a documentacin.
;           - Simplificar listas de push-t.
;
;


;
; Variables directorios
; ---------------------

(defvar path-tron
  (merge-pathnames
   (directory-namestring
    (eval (or (find-symbol "*LOAD-TRUENAME*")
	      (find-symbol "*LOAD-PATHNAME*")
	      (find-symbol "*LOAD-PATHNAME*" "SYSTEM") ; Para GCL
	      ""))))
  "Directorio que contiene los distintos ficheros necesarios para el juego Tron.")

(defvar path-tron-arena
  (if (directory (merge-pathnames (pathname "arena") path-tron))
      (merge-pathnames (pathname "arena/") path-tron)
    path-tron)
  "Directorio que contiene los distintos ficheros lisp de agentes jugadores de Tron.")

(defvar sender-tron-arena
  (merge-pathnames (pathname "sender-tron-arena") path-tron)
  "Programa encargado de enviar ficheros a arena.")

(defvar deleter-tron-arena
  (merge-pathnames (pathname "deleter-tron-arena") path-tron)
  "Programa encargado de eliminar ficheros de arena.")

(defconstant path-tron-sprod
  (if (directory (merge-pathnames (pathname "sprod") path-tron))
      (merge-pathnames (pathname "sprod/") path-tron)
    path-tron)
  "Directorio que contiene los packages para implementar sistemas de produccin y bsquedas.")


;
; Caraga de paquetes tron-base y tron-comm
; ----------------------------------------

(load (merge-pathnames (pathname "tron-base-package") path-tron))
(load (merge-pathnames (pathname "tron-comm-package") path-tron))
(load (merge-pathnames (pathname "tron-base") path-tron))
(load (merge-pathnames (pathname "tron-comm") path-tron))

(use-package "TRON-COMM")


;
; Caraga de paquetes shell, q_search y a_star
; -------------------------------------------

(defun load-if-exist (pathname)
  "Carga el fichero en PATHNAME si existe."
  (if (directory pathname)
      (load pathname)))

(load-if-exist (merge-pathnames (pathname "shell") path-tron-sprod))
(load-if-exist (merge-pathnames (pathname "q_search") path-tron-sprod))
(load-if-exist (merge-pathnames (pathname "a_star") path-tron-sprod))


;
; Variables version
; ------------------

(defvar *version-tron-loader* "0.5" "Versin de tron-loader.")

(defun version-tron-loader ()
  "Versin del tron-loader que se ha cargado."
  *version-tron-loader*)

(defun version-tron ()
  "Imprime las versin del los distintos mdulos Tron que se han cargado."
  (nconc
   `(("TRON-BASE" ,(tron-base:version-tron-base))
     ("TRON-COMM" ,(tron-comm:version-tron-comm))
     ("tron-loader" ,(version-tron-loader)))
   (mapcar (lambda (var-ag)
	     (let ((pn (package-name (symbol-package var-ag))))
	       (list pn
		     (eval (find-symbol (format nil "*VERSION-~A*" pn) pn)))))
	    lis-pkt-ag-load)))


;
; Variables globales
; ------------------

(defvar lis-lisp-types '(".lsp")
  "Lista de tipos de ficheros LISP ordenada por preferencias.")

(defvar lis-agentes nil
  "Lista de los agentes cargados en memoria actualmente.")
; Contiene listas con las funciones necesarias para cada agente:
;         funcin-movimiento: obligatoria
;         funcin-inicializacin: opcional
;         funcin-finalizacin: opcional
;         funcin-seguimiento-movimientos: opcional
;         funcin-seguimiento-cambios-direccin: opcional
;         funcin-seguimiento-consultas-contenido-posicion: opcional
;         funcin-seguimiento-consultas-posicion-jugador: opcional
;         funcin-seguimiento-consultas-direccin-jugador: opcional
;
;  Las 3 ltimas sin implementar en el entorno actualmente

(defvar lis-ag-vis nil
  "Lista de funciones de agentes visualizadores.")

(defvar lis-pkt-ag-load nil
  "Lista de packages de agentes cargados en memoria.")

(defvar lis-fich-load-arena nil
  "Lista de ficheros cargados en memoria del repositorio arena.")

(defvar lis-ag-hum nil
  "Lista de funciones de agente humano.")

(defvar *vis-tron* 1
  "Indica el nmero de agente visualizador a utilizar en partidas del juego Tron. NIL significa no utilizar agente visualizador.")

(defvar *dev-vis-tron* t
  "Indica el nombre del dispositivo a utilizar por agente visualizador en partidas del juego Tron para su salida, normalmente T  algo como \"/dev/pts/14\".")

(defvar *log-tron* nil
  "Indica si utilizar agente log en partidas del juego Tron.")

(defvar *dim-tab-tron* nil
  "Indica las dimensiones del tablero (x y) en partidas del juego Tron.")

(defvar *sin-bordes-tron* nil
  "Indica no usar bordes en el tablero en partidas del juego Tron.")

(defvar *sleep-cs-tron* nil
  "Indica las centsimas de segundo entre movimientos en partidas del juego Tron, NIL es el valor por defecto.")

(defvar *no-verbose-tron* nil
  "Indica desactivar verbose en partidas del juego Tron.")

(defvar *asincrono-tron* nil
  "Indica untilizar modo asncrono en partidas del juego Tron.")


; Especfico segn implementacin lisp
; ------------------------------------

(defun pref-punt (st)
  (concatenate 'string "." st))

(let ((lit (lisp-implementation-type)))
  (cond
   ((string-equal lit "GNU Common Lisp (GCL)")
    (setf lis-lisp-types
	  (eval (find-symbol "*LOAD-TYPES*" "SYSTEM"))))

  ((string-equal lit "CMU Common Lisp")
   (use-package "EXTENSIONS")
   (setf lis-lisp-types
	 (mapcar #'pref-punt
		 (append (eval (find-symbol "*LOAD-OBJECT-TYPES*"))
			 (eval (find-symbol "*LOAD-SOURCE-TYPES*"))))))

   ((string-equal lit "CLISP")
    (setf lis-lisp-types
	 (mapcar #'pref-punt
		 (append (eval (find-symbol "*COMPILED-FILE-TYPES*"))
			 (eval (find-symbol "*SOURCE-FILE-TYPES*"))))))
   (T ; Desconocida
    (my-warning "Implementacin ~S no soportada." lit))))


;
; Funciones de caga de agentes
; ----------------------------

(defun busca-ag (ag &optional (lista lis-agentes))
  "Busca agente AG en LISTA."
  (rest (assoc ag lista :test #'string=)))

(defun inserta-ag (ag)
  "Inserta agente AG en LIS-AGENTES."
  (let ((name-ag (symbol-name (first ag))))
    (if (busca-ag name-ag)
	(my-warning "Ya existe un agente con el nombre ~S en memoria." name-ag)
      (push (cons name-ag ag) lis-agentes))))

(defun load-tron-1 (file)
  "Carga y devuelve los agentes que se encuentren en el fichero de pathname FILE."
  (if (not (load file :verbose t :if-does-not-exist nil))
      (my-warning "Fichero ~S no se pudo cargar." file)
    (let (sym-lis-ag
	  (pkg (find-package (string-upcase (pathname-name file)))))
      (if (or (null pkg)
	      (null (setf sym-lis-ag 
			  (find-symbol "AGENTES" pkg))))
	  (my-warning "Fichero ~S no parece ser tipo agente tron." file)
	(progn
	  (push sym-lis-ag lis-pkt-ag-load)
	  ; Revisamos todos los agentes del fichero
	  (dolist (ag (eval sym-lis-ag) t)
	    (if (consp ag)
		(inserta-ag ag)
	      (inserta-ag (list ag)))))))))

(defun load-tron (file-str)
  "Carga los agentes que se encuentren en los ficheros emparejados por FILE-STR."
  (let (load-files)
    ; Recorremos todos los ficheros pasados
    (dolist (file (directory file-str) ; :truenamep nil)) ;en CMUCL
		  (reverse load-files))
      (if (load-tron-1 file)
	  (push file load-files)))))


(defun load-tron-arena (file-str)
  "Carga los agentes que se encuentren el los ficheros emparejados por FILE-STR dentro de repositorio arena."
  (let (load-files)
    ; Recorremos todos los ficheros pasados
    (dolist (file (directory (merge-pathnames (pathname file-str) path-tron-arena))
                    ; :truenamep nil)) ;en CMUCL
		  (reverse load-files))
      (if (member file lis-fich-load-arena :test #'equal)
	  (my-warning "Fichero ~S ya fue previamente cargado." file)
	(progn
	  (load-tron file)
	  (push file load-files)
	  (push file lis-fich-load-arena))))))


;
; Funciones de listado de agentes
; -------------------------------

(defun time-to-ymdhms (time)
  "Transforma de universal-time a lista de ao mes dia hora minuto y segundo."
  (multiple-value-bind (s m h d mt y)
      (decode-universal-time time)
    (list y mt d h m s)))

(defun ls-tron-arena (&optional ful (pat "*") &key full)
  "Devuelve los ficheros con agentes jugadores tron almacenados actualmente en la seccin arena. FULsi est a T devuelve un listado ampliado. PAT permite indicar un patron a emparejar."
  (let (lis-arena)
    (dolist (file (directory (merge-pathnames (pathname pat) path-tron-arena))
		  (reverse lis-arena))
      (if (or ful full)
	  (push (list (pathname-name file)
		      (file-author file)
		      (time-to-ymdhms (file-write-date file)))
		lis-arena)
	  (push (pathname-name file) lis-arena)))))
    
(defun datos-pkg (pkg)
  (let ((pkg-name (package-name pkg)))
    (first (ls-tron-arena t
	    (first (member pkg-name lis-fich-load-arena 
			   :test (lambda (s p)
				   (string-equal s (pathname-name p)))))))))

(defun jug-tron (&optional ful &key full)
  "Devuelve los agentes jugadores tron cargados actualmente en memoria. FUL si est a T devuelve un listado ampliado."
  (let (lis-fun-ag)
    (dolist (ag lis-agentes lis-fun-ag)
      (if (or ful full)
	  (push (list (first ag)
		      (datos-pkg (symbol-package (second ag)))
		      (documentation (second ag) 'function))
		lis-fun-ag)
	(push (first ag) lis-fun-ag)))))

(defun stat-tron-arena (&optional agentes agentes2 &rest options)
  "Muestra las estadisticas de partidas perdidas entre agentes tron de arena. Si se le pasa una lista AGENTES solo mostrar la lista con esos agentes. Si tambin se le pasa una lista AGENTES2 entonces enfrenta (tron-todos) los agentes de AGENTES contra los de AGENTES2, devolviendo y modificando 'hof.lsp' si es posible."
  (let* ((filename (merge-pathnames (pathname "hof.lsp") path-tron))
	 (st (open filename :direction :input :if-does-not-exist :error))
	 (lis-res (read st)))
    (close st)
    (if (not agentes2)
	(cond
	 ((null agentes) lis-res)
	 ((listp agentes)
	  (mapcar #'(lambda (ag) (cons ag (busca-ag ag lis-res)))
		  agentes))
	 (t (cons agentes (busca-ag agentes lis-res))))
      ; Modificar estadsticas aadiendo agentes
      (progn
	(setf lis-res
	      (apply #'tron-todos-internal (list agentes agentes2 lis-res) options))
	(setf st (ignore-errors (open filename :direction :output 
				      :if-does-not-exist :create
				      :if-exists :rename-and-delete)))
	(if st
	    (progn
	      (format st ";;;~%;;; Hall of fame de Tron~%;;; (no modificar manualmente)~%;;;~%")
	      (print lis-res st)
	      (close st)))))
	lis-res))

(defun nuevos-tron-arena (&optional agentes)
  "Indica los agentes, de los cargados actaulmente en memoria, que no se encuentran en 'hof.lsp'. Si AGENTES esta instanciado busca solo con esa lista."
  (let* ((lis-ag (if agentes (del-ag-humano agentes) 
		   (del-ag-humano (jug-tron))))
	 (filename (merge-pathnames (pathname "hof.lsp") path-tron))
	 (st (open filename :direction :input :if-does-not-exist :error))
	 (lis-res (read st)))
    (close st)
    (remove-if #'(lambda (ag)
		   (busca-ag ag lis-res))
	       lis-ag)))
    

;
; Funciones de transferenica de agentes
; -------------------------------------


(defun send-tron-arena (fich-agentes)
  "Enva un fichero de agentes Tron FICH-AGENTES (.lsp) al entorno arena."
  (if (string-equal (lisp-implementation-type) "CMU Common Lisp")
      (if (directory sender-tron-arena)
	  (let ((exited (run-program sender-tron-arena (list fich-agentes) :output t)))
	    (eq (process-exit-code exited) 0))
	(my-warning "No presente ~A." sender-tron-arena))
    (my-warning "send-tron-arena no implementado para ~A, usar CMUCL." (lisp-implementation-type))))


(defun del-tron-arena (fich-agentes)
  "Elimina el fichero de agentes Tron FICH-AGENTES (sin .lsp) del entorno arena. Ha de ser el propietario del fichero."
  (if (string-equal (lisp-implementation-type) "CMU Common Lisp")
     (if (directory deleter-tron-arena)
	  (let ((exited 
		 (run-program deleter-tron-arena (list fich-agentes)
			      :output t)))
	    (eq (process-exit-code exited) 0))
       (my-warning "No presente ~A." deleter-tron-arena))
    (my-warning "del-tron-arena no implementado para ~A, usar CMUCL." (lisp-implementation-type))))


;
; Funciones para lanzar partidas
; ------------------------------

(defmacro push-t (elm lis)
  `(if ,elm
       (push ,elm ,lis)))


(defun tron-internal (lis-jug
		   &key (vis *vis-tron*) ; (log nil)
		   (dev-vis *dev-vis-tron*)
		   num-jug
		   (dim-tab *dim-tab-tron*)
		   (sin-bordes *sin-bordes-tron*)
		   (sleep-cs *sleep-cs-tron*)
		   (no-verbose *no-verbose-tron*)
		   (asincrono *asincrono-tron*)
		   sleep-nil)
  "Inicia partida tron entre agentes pasados en LIS-JUG. La key VIS indica el nmero de agente-tron-vis que se  utilizar (1) o si no se utiliza (nil). DEV-VIS indica un dispositivo normalmente de terminal (/dev/pts/*) donde emitir la salida del agente visualizador seleccionado. El resto de keys definen el tipo de partida."
  (let (lis-fun-gen-dir
	lis-fun-ini
	lis-fun-fin
	lis-fun-seg-pos
	lis-fun-seg-dir
	lis-fun-seg-cposj
	lis-fun-seg-cestj
	lis-fun-seg-cpost
	(tipo-juego (list num-jug dim-tab sin-bordes sleep-cs no-verbose asincrono sleep-nil)))
    (if (null lis-jug)
        ; Por defecto juegan dos humanos
	(setf lis-jug (list (first lis-ag-hum) (first lis-ag-hum)))) 
    ; Revisamos que esten cargados agentes de jugadores pasados
    (dolist (ag lis-jug)
      (let ((lis-ag (busca-ag ag)))
	(if (null lis-ag)
	    (my-warning "Agente ~S no est cargado en memoria." ag)
	  (progn
	    (push (first lis-ag) lis-fun-gen-dir)
	    (push (second lis-ag) lis-fun-ini)
	    (push (third lis-ag) lis-fun-fin)
	    (push-t (fourth lis-ag) lis-fun-seg-pos)
	    (push-t (fifth lis-ag) lis-fun-seg-dir)
	    (push-t (sixth lis-ag) lis-fun-seg-cposj)
	    (push-t (seventh lis-ag) lis-fun-seg-cestj)
	    (push-t (eighth lis-ag) lis-fun-seg-cpost)))))

    (if (or (null lis-fun-gen-dir)
	    (and num-jug (not (eq num-jug (length lis-fun-gen-dir)))))
	(progn
	  (error "Nmero de jugadores no adecuado.")
	  nil)
      (progn
	(if (null num-jug)
	    (setf (first tipo-juego) (length lis-fun-gen-dir)))
	(if (numberp vis)
	    (let ((fun-vis (rest (nth vis lis-ag-vis))))
	      (push-t (second fun-vis) lis-fun-ini)
	      (push-t (third fun-vis) lis-fun-fin)
	      (push-t (fourth fun-vis) lis-fun-seg-pos)
	      (push-t (fifth fun-vis) lis-fun-seg-dir)
	      (push-t (sixth fun-vis) lis-fun-seg-cposj)
	      (push-t (seventh fun-vis) lis-fun-seg-cestj)
	      (push-t (eighth fun-vis) lis-fun-seg-cpost)
	      ; Abrir streamer para tron-vis
	      (let ((st-vis (find-symbol "*TRON-VIS-STREAM*" "TRON-VIS")))
		(if st-vis
		    (set st-vis
			 (cond
			   ((eq dev-vis t) *standard-output*)
			   ((eq dev-vis nil) nil)
			   (t
			    (open dev-vis :direction :output
				  :if-exists :overwrite))))))))

	(setf lis-fun-gen-dir (reverse lis-fun-gen-dir)
	      lis-fun-ini (reverse lis-fun-ini)
	      lis-fun-fin (reverse lis-fun-fin)
	      lis-fun-seg-pos (reverse lis-fun-seg-pos)
	      lis-fun-seg-dir (reverse lis-fun-seg-dir)
	      lis-fun-seg-cposj (reverse lis-fun-seg-cposj)
	      lis-fun-seg-cestj (reverse lis-fun-seg-cestj)
	      lis-fun-seg-cpost (reverse lis-fun-seg-cpost))

        ; Enviamos peticin de inicio de partida
	(envio-pet-ini lis-fun-gen-dir
		       lis-fun-ini
		       lis-fun-fin
		       lis-fun-seg-pos
		       lis-fun-seg-dir
		       lis-fun-seg-cposj
		       lis-fun-seg-cestj
		       lis-fun-seg-cpost
		       tipo-juego)))))


; Thanks to Barry Margolin:
; Subject: Re: &rest and &key;  group: comp.lang.lisp:
; http://groups.google.com/group/comp.lang.lisp/tree/browse_frm/thread/dcf71f6bdd232985/a9c3e3362c141ae2?rnum=1&q=%26rest+%26key+defun&_done=%2Fgroup%2Fcomp.lang.lisp%2Fbrowse_frm%2Fthread%2Fdcf71f6bdd232985%2Fa9c3e3362c141ae2%3Ftvc%3D1%26q%3D%26rest+%26key+defun%26#doc_4a9a5c342e93a827
    
(defun partition-args (keywords arglist) 
  "Partition the argument list around the first argument that is in KEYWORDS." 
  (let ((first-key (position-if #'(lambda (elt) (member elt keywords)) 
                                arglist)))
    (values (subseq arglist 0 first-key)
            (and first-key (subseq arglist first-key)))))

(defun tron (&rest arglist)
    "Inicia partida Tron entre agentes pasados. Para parmetros reales ver funcin TRON-INTERNAL."
  (multiple-value-bind (args options) 
      (partition-args '(:vis :dev-vis :num-jug :dim-tab :sin-bordes :sleep-cs :no-verbose :asincrono :sleep-nil) arglist)
    (apply #'tron-internal args options)))


(defun enfrenta-tron (ag1 ag2 arglist lis-res)
  "Enfrenta a dos agentes tron."
  (let* ((pareja (list ag1 ag2))
	 (res (apply #'tron-internal pareja arglist))
	 (stad1 (busca-ag ag1 lis-res))
	 (stad2 (busca-ag ag2 lis-res)))
    (setf (second stad1) (1+ (second stad1))
	  (second stad2) (1+ (second stad2)))
    (dolist (ag-pierde res)
      (let ((stad-per (busca-ag (nth (1- ag-pierde) pareja) lis-res)))
	(setf (first stad-per) (1+ (first stad-per)))))))

(defun del-ag-humano (list-ag)
  "Elimina AGENTE-HUMANO de la lista."
  (remove-if #'(lambda (ag)
	       (string= ag "AGENTE-HUMANO"))
	     list-ag))

(defun tron-todos-internal (args &rest arglist)
  "Compite agentes en memoria todos contra todos dos a dos, emitiendo una lista final con sublistas conteniendo: nombre del agente, nmero de partidas perdidas y nmero de partidas jugadas. El resultado final es ordenado por nmero de partidas perdidas dividido entre nmero de partidas jugadas."
  (let* ((lis-jug-1 (if (first args) (del-ag-humano (first args))
			(del-ag-humano (jug-tron))))
	 (lis-jug-2 (if (second args) (del-ag-humano (second args))
			(del-ag-humano (jug-tron))))
	 (lis-res (third args)))
         ; Lista asociativa de resultados
    (dolist (ag (append lis-jug-1 lis-jug-2))
      (if (null (busca-ag ag lis-res))
	  (push (list ag 0 0) lis-res)))
    ; Enfrentamientos restringidos
    (dolist (ag1 lis-jug-1)
      (dolist (ag2 lis-jug-2 (progn (print lis-res)
				    (force-output)))
	(enfrenta-tron ag1 ag2 arglist lis-res)))
    (dolist (ag lis-jug-1)
      (setf lis-jug-2 (remove ag lis-jug-2 :test #'string=)))
    (dolist (ag1 lis-jug-2)
      (dolist (ag2 lis-jug-1 (progn (print lis-res)
				    (force-output)))
	(enfrenta-tron ag1 ag2 arglist lis-res)))
    (sort lis-res #'<
	  :key #'(lambda (s)
		   (/ (second s) (third s))))))


(defun tron-todos (&rest arglist)
  "Compite agentes en memoria todos contra todos dos a dos, emitiendo una lista final con sublistas conteniendo: nombre del agente, nmero de partidas perdidas y nmero de partidas jugadas. El resultado final es ordenado por nmero de partidas perdidas dividido entre nmero de partidas jugadas. Los parmetros que se le pasan realmente son (&OPTIONAL LIST-AG-1 LIST-AG-2 &KEY TRON-KEYS), donde LIST-AG-1 y LIST-AG-2 son listas de nombres agentes a enfrentar entre s, si no se pasa se inicializan con el resultado de (JUG-TRON). TRON-KEYS son las keywords a pasar a la funcin TRON-INTERNAL."
  (multiple-value-bind (args options) 
      (partition-args '(:vis :dev-vis :num-jug :dim-tab :sin-bordes :sleep-cs :no-verbose :asincrono :sleep-nil) arglist)
    (apply #'tron-todos-internal args options)))

  

;
; Inicializacin
; --------------

(load-tron-1 (merge-pathnames (pathname "tron-vis") path-tron))
(setf lis-ag-vis (reverse lis-agentes))
(setf lis-agentes nil)

; (load-tron (concatenate 'string path-trho "tron-log"))
;(setf lis-ag-log (rest (first lis-agentes)))
;(setf lis-agentes nil)

(load-tron-arena "tron-agentes-humanos")
(setf lis-ag-hum (first lis-agentes))

(load-tron-arena "tron-agentes-simples")
