; Programa: Package para agentes visualizadores del juego Tron,
;           juego tambin denominado "light cycle",
;           ver http://en.wikipedia.org/wiki/Light_cycle
;
; Asignatura: Introduccin a la Inteligencia Artificial
;             ETSII - ULL
; Prctica: (sistemas de produccin)
; Profesor: Patricio Garca (patricio AT etsii.ull.es)
; Fecha:    12/04/2006
; Comentarios:
;
;           Las funciones accesibles ms importantes son:
;
;           Ver la documentacin existente para cada una de ellas.
;
;           Para su utilizacin es aconsejable usar el cargador
;           del entorno de juego tron: tron-loader.lsp.
;
;
; Dependencias:
;           Hace uso del paquete que contiene la interface (tron-comm)
;           con el control del juego (tron-base).
;
; Modificaciones:
;
;           - Ver 0.1 (13/04/2006): Primera versin. Extraida
;             de tron.0.1.lsp. Vase tron-base.lsp y tron-comm*.lsp.
;             Realizada ocultacin del cursor. Incluido tiempo de
;             espera al final. Arreglado visulaizacin de
;             descripcin de funciones agentes y nombres.
;
;           - Ver 0.1.1 (03/05/2006): Reparado bug que desplazaba
;             bordes del tablero.
;
;           - Ver 0.1.2 (03/05/2006): Cambio en fin-vis-tron debido
;             a cambio en tron-comm de ver "fun-0.2".
;
;           - Ver 0.2 (05/05/2006): Inclusin de variable
;             *tron-vis-stream* que permite visualizar el tablero
;             en otro streamer, ej (open "/dev/pts/12" :direction 
;             :output :if-exists :overwrite). Buscarla con "
;             w $LOGNAME | grep "w $LOGNAME"". Aadido versin
;             del package en el borde superior izquierdo.
;
;           - Ver 0.3 (12/05/2006): Incluida funciones para
;             el seguimiento de consultas, debido a cambios
;             en tron-comm de ver "fun-0.3". Generados varios
;             visualizadores segn seguimientos. Eliminado bug
;             que no limpiaba bien la lnea de estado
;             (con "/dev/pts/...").
;
;           - Ver 0.3.1 (20/05/2006): Corregido bug en 
;             borra-pinta-seg que impeda pintar "". Cambiado
;             seguimiento para que pinte en cuanto se recibe.
;
;           - Ver 0.3.2 (27/05/2006): Aadido envio-punteo-pos.
;             Aadido envio-mensaje.
;
; Por hacer:
;           - Correccin de Bugs
;             ------------------
;           - interpreted-function ha de ser eval:interpreted-function,
;             pero esto provoca un error en el gcl de paquete 
;             no encontrado.
;           - Ver como conjuntar envio-refresco-tab en interface de comm.
;           - definir *version-tron-vis* como constante?
;           - A veces aparece: ** Closed file "/dev/pts/14"
;
;
;           - Mejoras en funcionalidades
;             --------------------------
;           - Verificar funcionamiento para <> de 2 jugadores.
;             Problemas caracteres, colores, salto lnea status.
;           - Tratar de independizar ms la visualizacin.
;           - Revisar nombre de variables.
;           - Agregar la documentacin a cada funcin importante.
;           - Que vis-tron y vis-jug-tron impriman ! en el mismo color?
;           - Aadir aqu, o en agente-humano, un agente que permita
;             "pausar" la ejecucin?
;           - Hacer versin con gcl-tk.
;           - Revisar tipos de visualizadores.
;           - Aadir posiblidad de recibir e imprimir mensaje.
;           - utilizar el title bar de xterm?
;
;




;
; Definicin del package
; ----------------------

(defpackage "TRON-VIS"
  (:nicknames "TRON-VIS-XTERM" "tr-vis" "tron-v")
  (:use "LISP" "TRON-COMM")
  (:documentation "Agentes visualizadores de partidas Tron")
  (:export "AGENTES" "VERSION-TRON-VIS" "VIS-BAJO" "VIS-NORMAL" "VIS-CPOST" "VIS-CDIRJ" "VIS-CPOSJ" "VIS-FULL" "*TRON-VIS-STREAM*" "INI-VIS-TRON" "FIN-VIS-TRON" "FIN-VIS-TRON0" "SEG-POS-VIS-TRON" "SEG-DIR-VIS-TRON" "SEG-CPOSJ-VIS-TRON" "SEG-CDIRJ-VIS-TRON" "SEG-CPOST-VIS-TRON" "ENVIO-REFRESCO-TAB" "ENVIO-PUNTEO-POS" "ENVIO-MENSAJE"))

(in-package "TRON-VIS")


;
; Constante con lista de agentes presentes en este package
; --------------------------------------------------------

(defvar agentes '((vis-bajo vis-bajo fin-vis-tron0)
		  (vis-normal vis-normal fin-vis-tron seg-pos-vis-tron)
		  (vis-cpost vis-cpost fin-vis-tron seg-pos-vis-tron nil nil nil seg-cpost-vis-tron)
		  (vis-cdirj vis-cdirj fin-vis-tron seg-pos-vis-tron nil nil seg-cdirj-vis-tron seg-cpost-vis-tron)
		  (vis-cposj vis-cposj fin-vis-tron seg-pos-vis-tron nil seg-cposj-vis-tron nil seg-cpost-vis-tron)
		  (vis-full vis-full fin-vis-tron seg-pos-vis-tron nil seg-cposj-vis-tron seg-cdirj-vis-tron seg-cpost-vis-tron))
  "Lista de funciones de agente de visualizacin")

;
; Variables version
; ------------------

(defvar *version-tron-vis* "xterm-0.3.2" "Versin del package tron-vis.")

(defun version-tron-vis ()
  "Devuelve la versin del tron-vis que se ha cargado."
  *version-tron-vis*)

;
; Variables globales
; ------------------

(defvar *tron-vis-stream* *standard-output*
  "Stream de visualizacin del tablero Tron. Puede ser un fichero de terminal (/dev/pts/*).")

(defvar num-jug-tron 2
  "Numero de jugadores en la partida actual.")

(defvar pos-jug-tron nil
  "Array para almacenar las posiciones actuales de cada jugador.")

(defun pos-jug (jug)
  "Clculo local de la posicin del jugador JUG."
  (copy-list (aref pos-jug-tron (1- jug))))

(defvar msg-jug-tron nil
  "Array para almacenar los mensajes recibidos de cada jugador")

(defvar sleep-cs-tron 0
  "Tiempo en centsimas de segundo entre turnos.")

(defvar dim-tab-tron nil
  "Dimensiones del tablero Tron.")

(defvar status-pos '(0 23)
  "Posicin (X Y) de comienzo de la lnea de estado. Debajo del tablero.")

(defun calc-status-pos ()
  "Calculo de posicin (X Y) de comienzo de la lnea de estado. Debajo del tablero."
  (list 0 (second dim-tab-tron)))

(defvar nom-vis ""
  "Nombre del visualizador en uso.")

(defvar doc-vis ""
  "Documentacin del visualizador en uso.")

;
; Funciones nombres de agentes visualizadores
; -------------------------------------------

(defun vis-bajo (&rest param)
  "Visualizador ms simple, solo muestra situacin al inicio y al final de la partida."
  (setf nom-vis "VIS-BAJO"
	doc-vis (documentation 'vis-bajo 'function))
  (apply #'ini-vis-tron param))

(defun vis-normal (&rest param)
  "Visualizador por defecto, no realiza seguimientos de las consultas de los agentes."
  (setf nom-vis "VIS-NORMAL"
	doc-vis (documentation 'vis-normal 'function))
  (apply #'ini-vis-tron param))

(defun vis-cpost (&rest param)
  "Visualizador que realiza seguimiento de las consultas sobre el contenido de posiciones del tablero (consulta-pos-tab): caracter \"\" para consultas absolutas o \"^\", \">\" \"v\" y \"<\" para consultas relativas."
  (setf nom-vis "VIS-CPOST"
	doc-vis (documentation 'vis-cpost 'function))
  (apply #'ini-vis-tron param))

(defun vis-cdirj (&rest param)
  "Visualizador que realiza seguimiento de las consultas sobre la direccin de los agentes, caracter \"?\", y el contenido de posiciones del tablero: caracter \"\" para consultas absolutas o \"^\", \">\" \"v\" y \"<\" para consultas relativas."
  (setf nom-vis "VIS-CDIRJ"
	doc-vis (documentation 'vis-cdirj 'function))
  (apply #'ini-vis-tron param))

(defun vis-cposj (&rest param)
  "Visualizador que realiza seguimiento de las consultas sobre la posicin de los agentes, caracter \":\", y el contenido de posiciones del tablero: caracter \"\" para consultas absolutas o \"^\", \">\" \"v\" y \"<\" para consultas relativas."
  (setf nom-vis "VIS-CPOSJ"
	doc-vis (documentation 'vis-cposj 'function))
  (apply #'ini-vis-tron param))

(defun vis-full (&rest param)
  "Visualizador que realiza todos los seguimientos: consultas sobre la posicin de los agentes, caracter \":\", la direccin de los agentes, caracter \"?\", y el contenido de posiciones del tablero: caracter \"\" para consultas absolutas o \"^\", \">\" \"v\" y \"<\" para consultas relativas."
  (setf nom-vis "VIS-FULL"
	doc-vis (documentation 'vis-full 'function))
  (apply #'ini-vis-tron param))


;
; Funciones especficas segn implementacin lisp
; -----------------------------------------------

(let ((lit (lisp-implementation-type)))
  (cond
   ((string-equal lit "GNU Common Lisp (GCL)"))
   ((string-equal lit "CMU Common Lisp")
    (set (find-symbol "*GC-VERBOSE*" "EXTENSIONS") nil))
   ((string-equal lit "CLISP"))
   (T ; Implementacin desconocida
    (my-warning "Implementacin ~S no soportada" lit))))


;
; Primitivas para acceso a pantalla
; ---------------------------------

(defconstant escape-char (code-char 27)
  "Escape char.")

(defconstant command-cursor 
  (concatenate 'string (string escape-char) "[")
  "String de envio de comandos al cursor.")

(defconstant command-cursor-color-ini 
  (concatenate 'string (string escape-char) "[")
  "String de envio de incicio de comando de color al cursor.")

(defconstant command-cursor-color-end "m"
  "String de envio de final de comando de color al cursor.")

(defconstant cursor-color-nothing "0")
(defconstant cursor-color-light-purple "1;35")
(defconstant cursor-color-light-blue "1;34")
(defconstant cursor-color-light-green "1;32")

(defconstant command-title-bar-ini
  (concatenate 'string (string escape-char) "]0;")
  "String de envio de incicio de comando al title bar.")

(defconstant command-title-bar-end (code-char 7)
  "Caracter de envio de final de comando al title bar.")


(defun set-cursor-off (&optional (st *standard-output*))
  "Oculta el cursor."
  (princ command-cursor st)
  (princ "?25l" st))

(defun set-cursor-on (&optional (st *standard-output*))
  "Muestra el cursor."
  (princ command-cursor st)
  (princ "?25h" st))

(defun save-window-cursor-position (&optional (st *standard-output*))
  "Salva la posicin del cursor."
  (princ command-cursor st) (princ #\s st))

(defun restore-window-cursor-position (&optional (st *standard-output*))
  "Recupera la posicin del cursor salvada."
  (princ command-cursor st) (princ #\u st))

(defun set-window-cursor-position (line column &optional (st *standard-output*))
  "Posiciona el cursor."
  (princ command-cursor st)
  (format st "~D;~DH" line column))

(defun set-window-cursor-color (color &optional (st *standard-output*))
  "Establecer color del cursor."
  (princ command-cursor-color-ini st)
  (format st "~A~A" color command-cursor-color-end))

(defun set-title-bar (msg &optional (st *standard-output*))
  "Imprime mensaje MSG en title-bar."
  (princ command-title-bar-ini st)
  (princ msg st)
  (princ command-title-bar-end st)
  nil)

;
; Visualizacin de estructura de problema
; ---------------------------------------

(defvar lis-ags-jug nil
  "Lista de agentes jugadores")

(defvar char-jug1 #\*
  "Caracter para visualizar el jugador 1")

(defvar char-jug2 #\+
  "Caracter para visualizar el jugador 2")

(defvar char-obs #\#
  "Caracter para visualizar los obstculos")

(defvar char-end #\!
  "Caracter para visualizar el choque final")

(defvar color-jug1 cursor-color-light-purple
  "Color para visualizar el jugador 1")

(defvar color-jug2 cursor-color-light-blue
  "Color para visualizar el jugador 2")

(defvar color-obs cursor-color-nothing
  "Color para visualizar los obstculos")

(defvar color-end cursor-color-light-green
  "Color para visualizar el choque final")

(defvar elm-mat-char-color
  (make-array '(4 2) :initial-contents
	      `((,char-obs  ,color-obs)
		(,char-jug1 ,color-jug1)
		(,char-jug2 ,color-jug2)
		(,char-end  ,color-end)))
  "Array para transformar el contenido de elemento de tron-mat a carcter")

(defvar lis-con-seg nil
  "Lista de consultas realizadas pendientes de pintar.")

(defvar lis-con-pin nil
  "Lista de consultas realizadas ya pintadas pendientes de borrar.")

(defvar cpost-no-vis t
  "Indica si no es el propio visualizador el que realiza las consulta al tablero para que no se pinte.")

(defun vis-pos-color (posxy color obj &optional (st *standard-output*))
  "Posiciona el cursor en la posicin POSXY, si POSXY es no NIL, activa le color COLOR e imprime el objeto OBJ"
  (if posxy
      (set-window-cursor-position (1+ (second posxy)) (1+ (first posxy)) st))
  (set-window-cursor-color color st)
  (princ obj st))


(defun vis-tron-prev (nom-vis doc-vis &optional (st *standard-output*))
  "Visualizacin de informacin previa a partida TRON"
  (dotimes (i (length lis-ags-jug))
    (format st "~%Jugador ~D: ~A~%" (1+ i) (first (nth i lis-ags-jug)))
    (format st "~A~%" (second (nth i lis-ags-jug))))
  (format st "~%~A-~A: ~A~%" 
	  (package-name (symbol-package '*version-tron-vis*))
	  *version-tron-vis*
	  nom-vis)
  (format st "~A~%" doc-vis)
  (sleep (ceiling (* 12 sleep-cs-tron) 100))
  (format st "~%Comienza la partida ")
  (if (equal *standard-output* *tron-vis-stream*)
      (format st "...~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%")
    (format st "en ~S ...~%" *tron-vis-stream*))
  (force-output st))


(defun clear-status (&optional (st *standard-output*))
  "Se posiciona en la lnea de estado y la borra, dejando el cursor al principio de ella."
  (vis-pos-color status-pos cursor-color-nothing "" st)
  (dotimes (i (first dim-tab-tron))
    (princ #\Space st))
  (vis-pos-color status-pos cursor-color-nothing "" st))
  

(defun print-nom-ag (&optional list-chocados (st *standard-output*))
  "Imprime nombres de agentes en la barra de estado."
  (clear-status st)
  (dotimes (i num-jug-tron)
    (let ((jug (1+ i))
	  (agent-name (first (nth i lis-ags-jug))))
      (vis-pos-color nil 
		     (aref elm-mat-char-color (rem jug 4) 1)
		     (format nil "~D" jug)
		     st)
      (if (member jug list-chocados)
	  (progn
	    (vis-pos-color nil color-end char-end st)
	    (vis-pos-color nil 
		     (aref elm-mat-char-color (rem jug 4) 1)
		     (format nil "~C:~A "
			     (aref elm-mat-char-color (rem jug 4) 0)
			     agent-name)
		     st))
	(format st ":~C:~A "
			     (aref elm-mat-char-color (rem jug 4) 0)
			     agent-name)))))

(defun vis-tron (tab &optional (st *standard-output*))
  "Visualiza el tablero TAB utlizando caracteres. TAB ha de ser una matriz tipo TRON-MAT."
  (set-window-cursor-position 1 1 st)
  (dotimes (y (array-dimension tab 1) t)
    (dotimes (x (array-dimension tab 0) (terpri st))
      (let ((elem (aref tab x y)))
	(if elem
	    (progn
	      (set-window-cursor-color (aref elm-mat-char-color (rem elem 4) 1) st)
	      (princ (aref elm-mat-char-color (rem elem 4) 0) st))
	  (princ " " st)))))
  ; Imprimimos versin de tron-vis
  (set-window-cursor-position 1 1 st)
  (format st "~A-~A" 
	  (package-name (symbol-package '*version-tron-vis*))
	  *version-tron-vis*)
  ; Imprimimos los nombres de los agentes
  (print-nom-ag nil st)
  ; Luego nos posicionamos en la lnea de estado
  (vis-pos-color status-pos cursor-color-nothing "" st)
  (force-output st)
  t)

(defmacro col-jug (jug)
  `(if ,jug
       (aref elm-mat-char-color (rem ,jug 4) 1)
     color-obs))

(defmacro car-jug (jug)
  `(if ,jug
       (aref elm-mat-char-color (rem ,jug 4) 0)
     #\Space))

(defun vis-jug-tron (jug pos &optional (st *standard-output*))
  "Visualiza el jugador JUG del tablero en la posicin POS utlizando caracteres y colores."
  (vis-pos-color pos
		 (col-jug jug)
		 (car-jug jug)
		 st)
  t)

(defun vis-jug-end (est &optional (st *standard-output*))
  "Visualiza la finalizacin del juego, un ! en el choque. EST indica el estado de finalizacin del juego, ver funcin TRON-BASE:CHECK-SET-INC-POS."
  ; Escribimos caracteres de final
  (dotimes (i num-jug-tron)
    (let* ((jug (1+ i))
	   (posxy (consulta-pos-jug jug)))
      (if (member jug est)
	  (vis-pos-color posxy color-end char-end st)
	(vis-pos-color posxy
		       (aref elm-mat-char-color (rem jug 4) 1)
		       (aref elm-mat-char-color (rem jug 4) 0)
		       st))))
  ; Imprimimos los nombres de los agentes
  (print-nom-ag est st)
  ; Nos posicionamos en la lnea de estado
  (vis-pos-color status-pos cursor-color-nothing "" st)
  (set-cursor-on st)
  (force-output st))

(defun vis-pos-tron (pos &optional (st *standard-output*))
  "Visualiza (borra) una determinada posicin POS del tablero Tron."
  (setf cpost-no-vis nil)
  (vis-jug-tron (consulta-pos-tab pos) pos st)
  (setf cpost-no-vis t))

; dir-inc, dir2dim-inc, dir2val-inc y pos-inc-dir
; Extraidas de tron-base
; ----------------------

(defconstant dir-inc
  (make-array '(4 2) :initial-contents
	      '((1 -1)    ; Norte
		(0 +1)    ; Este
		(1 +1)    ; Sur
		(0 -1)))  ; Oeste
  "Array para transformar de direccin a campo dir-jug de tron")

(defmacro dir2dim-inc (dir)
  `(aref dir-inc ,dir 0))

(defmacro dir2val-inc (dir)
  `(aref dir-inc ,dir 1))

(defun pos-inc-dir (pos dir)
  "Calcula la posicin resultante de desplazarse en la direccinb DIR respecto a POS."
  (let ((pos-inc (copy-list pos)))
    (setf (nth (dir2dim-inc dir) pos-inc)
	  (+ (nth (dir2dim-inc dir) pos-inc)
	     (dir2val-inc dir)))
    pos-inc))

(defun pinta-seg (seg &optional (st *standard-output*))
  "Pinta un seguimiento."
  (case (first seg)
    (cposj (vis-pos-color (fourth seg) (col-jug (second seg)) #\: st))
    (cdirj (vis-pos-color (fourth seg) (col-jug (second seg)) #\? st))
    (cpost
     (if (fourth seg)
	 (setf (third seg) (pos-inc-dir (third seg) (fourth seg))))
     (vis-pos-color (third seg)
		    (col-jug (second seg))
		    (case (fourth seg)
		      (0 #\^)
		      (1 #\>)
		      (2 #\v)
			(3 #\<)
			(otherwise #\))
		    st))
    (cpunt (vis-pos-color (third seg) (col-jug (second seg)) #\- st)))
  (force-output st)
  seg)

(defun borra-segs (&optional (st *standard-output*))
  "Borra los seguimientos actualmente pintados."
  (dolist (elm lis-con-seg)
    (case (first elm)
      ((cposj cdirj) (vis-jug-tron (third elm) (fourth elm) st))
      ((cpost cpunt) (vis-pos-tron (third elm) *tron-vis-stream*))))
  (setf lis-con-seg nil))

(defun borra-pinta-seg (&optional (st *standard-output*))
  "No usada..Borra los seguimientos actualmente pintados y pinta los nuevos."
  ; Borramos consultas las actualmente pintadas
  (dolist (elm lis-con-pin)
    (case (first elm)
      ((cposj cdirj) (vis-jug-tron (third elm) (fourth elm) st))
      (cpost (vis-pos-tron (third elm) *tron-vis-stream*))))
  ; Pintamos las consultas nuevas
  (dolist (elm (reverse lis-con-seg))
    (case (first elm)
      (cposj (vis-pos-color (fourth elm) (col-jug (second elm)) #\: st))
      (cdirj (vis-pos-color (fourth elm) (col-jug (second elm)) #\? st))
      (cpost
       (if (fourth elm)
	   (setf (third elm) (pos-inc-dir (third elm) (fourth elm))))
       (vis-pos-color (third elm)
		      (col-jug (second elm))
		      (case (fourth elm)
			(0 #\^)
			(1 #\>)
			(2 #\v)
			(3 #\<)
			(otherwise #\))
			st))))
  ; Actualizamos listas
  (setf lis-con-pin lis-con-seg)
  (setf lis-con-seg nil))


;
; Interface de consultas tron-comm ==> agente-visualizacin
; ---------------------------------------------------------

(defun ini-vis-tron (jug tipo-juego ags-jug)
  "Funcin de inicializacin de partida."
  (declare (ignore jug))
  ; Asigna variables globales
  (setf lis-ags-jug ags-jug)
  (setf num-jug-tron (first tipo-juego))
  (setf sleep-cs-tron (fourth tipo-juego))
  (setf dim-tab-tron (second tipo-juego))
  (setf status-pos (calc-status-pos))
  (setf pos-jug-tron (make-array num-jug-tron :initial-element nil))
  (setf msg-jug-tron (make-array num-jug-tron :initial-element ""))
  (dotimes (i num-jug-tron)
    (setf (aref pos-jug-tron i) (consulta-pos-jug (1+ i))))
  (setf lis-con-seg nil
	lis-con-pin nil)
  ; Visualiza mensajes previos
  (vis-tron-prev nom-vis doc-vis)
  ; Visualiza tablero completo
  (set-cursor-off *tron-vis-stream*)
  (vis-tron (consulta-tab) *tron-vis-stream*)
  t)

(defun fin-vis-tron (jug est)
  "Funcin de finalizacin de partida."
  (declare (ignore jug))
  ;(borra-pinta-seg *tron-vis-stream*)
  (borra-segs *tron-vis-stream*)
  (vis-jug-end est *tron-vis-stream*)
  (sleep (ceiling (* 12 sleep-cs-tron) 100))
  (setf lis-ags-jug nil)
  t)

(defun fin-vis-tron0 (jug est)
  "Funcin de finalizacin de partida para agente vis-bajo."
  (vis-tron (consulta-tab) *tron-vis-stream*)
  (fin-vis-tron jug est))

(defun seg-pos-vis-tron (jug pos)
  "Funcin de seguimiento de cambio de posicin de jugador."
  (vis-jug-tron jug pos *tron-vis-stream*)
  (setf (aref pos-jug-tron (1- jug)) pos)
  (if (eq jug num-jug-tron)
      ;(borra-pinta-seg *tron-vis-stream*))
      (borra-segs *tron-vis-stream*))
  (force-output *tron-vis-stream*)
  t)

(defun seg-dir-vis-tron (jug dir)
  "Funcin de seguimiento de cambio de direccin de jugador. No implementada."
  (declare (ignore jug dir))
  nil)

(defmacro pinta-push-seg (lis-seg)
  "Pinta y apila seguimiento para luego borrar."
  `(push (pinta-seg ,lis-seg *tron-vis-stream*) lis-con-seg))

(defun seg-cposj-vis-tron  (jug-pet jug)
  "Funcin de seguimiento de consultas de posicin de jugador."
  (pinta-push-seg (list 'cposj jug-pet jug (pos-jug jug))))

(defun seg-cdirj-vis-tron  (jug-pet jug)
  "Funcin de seguimiento de consultas de direccin de jugador."
  (pinta-push-seg (list 'cdirj jug-pet jug (pos-jug jug))))

(defun seg-cpost-vis-tron  (jug-pet pos dir)
  "Funcin de seguimiento de consultas de posicin de tablero."
  (if cpost-no-vis
      (pinta-push-seg  (list 'cpost jug-pet pos dir))))


;
; Interface de consultas agentes-jugadores ==> agente-visualizacin
; -----------------------------------------------------------------

(defun envio-refresco-tab ()
  "Solicita refresco del tablero."
  (vis-tron (consulta-tab) *tron-vis-stream*)
  t)

(defun envio-punteo-pos (pos)
  "Solicita pintar punteo con '-' de posicin POS."
  (if lis-ags-jug
      (pinta-push-seg (list 'cpunt nil pos))))

(defun envio-mensaje (msg)
  "Solicita imprimir mensaje en tablero (title-bar)."
  (let ((jug (consulta-turno))
	(msg-tot ""))
    (if (and lis-ags-jug jug)
	(progn
	  (setf (aref msg-jug-tron (1- jug)) msg)
	  (dotimes (i num-jug-tron)
	    (if (string/= "" (aref msg-jug-tron i))
		(setf msg-tot
		      (concatenate 'string msg-tot
				   (format nil "~D: ~A "
					   (1+ i)
					   (aref msg-jug-tron i))))))
	  (set-title-bar msg-tot *tron-vis-stream*)))))

